# -*- coding: utf-8 -*-
"""
Script to cache anonymous houseprint data into hp_anonymous.pkl

Created on 05/07/2014 by Roel De Coninck
"""
import os

from opengrid.library.houseprint import houseprint
from opengrid import config
c = config.Config()


##############################################################################

hp = houseprint.Houseprint()
print('Sensor data fetched')

filename = os.path.join(c.get('data', 'folder'), 'hp_anonymous.pkl')
hp.save(filename)



__author__ = 'Jan Pecinovsky'

import pandas as pd

"""
A Device is an entity that can contain multiple sensors.
The generic Device class can be inherited by a specific device class, eg.
Fluksometer
"""


class Device(object):
    def __init__(self, key=None, site=None):
        self.key = key
        self.site = site
        self.sensors = []

    def __repr__(self):
        return """
    {}
    Key: {}
    {} sensors
    """.format(self.__class__.__name__,
               self.key,
               len(self.sensors)
              )

    def get_sensors(self, sensortype = None):
        """
            Return a list with all sensors in this Device

            Parameters
            ----------
            sensortype: gas, water, electricity: optional

            Returns
            -------
            list of Sensors
        """
        return [sensor for sensor in self.sensors if sensor.type == sensortype or sensortype is None]

    def get_data(self, sensortype=None, head=None, tail=None, diff='default', resample='min', unit='default'):
        """
        Return a Pandas Dataframe with the joined data for all sensors in this device

        Parameters
        ----------
        sensors : list of Sensor objects
            If None, use sensortype to make a selection
        sensortype : string (optional)
            gas, water, electricity. If None, and Sensors = None,
            all available sensors in the houseprint are fetched
        head, tail: timestamps,
        diff : bool or 'default'
            If True, the original data will be differentiated
            If 'default', the sensor will decide: if it has the attribute
            cumulative==True, the data will be differentiated.
        resample : str (default='min')
            Sampling rate, if any.  Use 'raw' if no resampling.
        unit : str , default='default'
            String representation of the target unit, eg m**3/h, kW, ...

        Returns
        -------
        Pandas DataFrame
        """

        sensors = self.get_sensors(sensortype)
        series = [sensor.get_data(head=head, tail=tail, diff=diff, resample=resample, unit=unit) for sensor in sensors]

        # workaround for https://github.com/pandas-dev/pandas/issues/12985
        series = [s for s in series if not s.empty]

        if series:
            df = pd.concat(series, axis=1)
        else:
            df = pd.DataFrame()

        # Add unit as string to each series in the df.  This is not persistent: the attribute unit will get
        # lost when doing operations with df, but at least it can be checked once.
        for s in series:
            try:
                df[s.name].unit = s.unit
            except:
                pass

        return df

    def number_of_sensors(self, sensortype=None):
        """

        Parameters
        ----------
        sensortype: gas, water, electricity

        Returns
        -------
        int
        """
        return len(self.get_sensors(sensortype=sensortype))

    def last_timestamp(self, epoch=False):
        """
        Get the last timestamp for a device, by returning the latest timestamp
        of the sensors

        Parameters
        ----------
        epoch : bool
            default False
            If True return as epoch
            If False return as pd.Timestamp

        Returns
        -------
        pd.Timestamp | int
        """
        timestamps = [sensor.last_timestamp(epoch=epoch) for sensor in self.sensors]
        return max(timestamps)

    def add_sensor(self, sensor):
        """
        Parameters
        ----------
        sensor : Sensor
        """
        sensor.device = self
        self.sensors.append(sensor)


class Fluksometer(Device):
    def __init__(self, site=None, key=None, mastertoken=None, tmpos=None):

        # invoke init method of generic Device
        super(Fluksometer, self).__init__(key, site)

        self.mastertoken = mastertoken
        self._tmpos = tmpos

    @property
    def tmpos(self):
        if self._tmpos is not None:
            return self._tmpos
        elif self.site.tmpos is not None:
            return self.site.tmpos
        else:
            raise AttributeError('TMPO session not defined')



__author__ = 'Jan Pecinovsky'

from opengrid.config import Config

config = Config()

import os
import sys
import json
import jsonpickle
import datetime as dt
import pandas as pd
from requests.exceptions import HTTPError
import warnings
from tqdm import tqdm

# compatibility with py3
if sys.version_info.major >= 3:
    import pickle
else:
    import cPickle as pickle

import tmpo

# compatibility with py3
if sys.version_info.major >= 3:
    from .site import Site
    from .device import Device, Fluksometer
    from .sensor import Sensor, Fluksosensor
else:
    from site import Site
    from device import Device, Fluksometer
    from sensor import Sensor, Fluksosensor

"""
The Houseprint is a Singleton object which contains all metadata for sites, devices and sensors.
It can be pickled, saved and passed around
"""


class Houseprint(object):
    def __init__(self,
                 gjson=None,
                 spreadsheet="Opengrid houseprint (Responses)",
                 empty_init=False
                 ):
        """
            Parameters
            ---------
            gjson: Path to authentication json
            spreadsheet: String, name of the spreadsheet containing the metadata
        """

        self.sites = []
        self.timestamp = dt.datetime.utcnow()  # Add a timestamp upon creation

        if not empty_init:
            if gjson is None:
                gjson = config.get('houseprint', 'json')
            self.gjson = gjson
            self.spreadsheet = spreadsheet
            self._parse_sheet()

    def reset(self):
        """
        Connect to the Google Spreadsheet again and re-parse the data
        """
        self.__init__(gjson=self.gjson, spreadsheet=self.spreadsheet)
        if hasattr(self, '_tmpos'):
            self._add_sensors_to_tmpos()  

    def __repr__(self):
        return """
    Houseprint
    Created on {} (UTC)
    {} sites
    {} devices
    {} sensors
    """.format(self.timestamp,
               len(self.sites),
               sum([len(site.devices) for site in self.sites]),
               sum([len(site.sensors) for site in self.sites])
               )

    def _parse_sheet(self):
        """
            Connects to Google, fetches the spreadsheet and parses the content
        """

        import gspread
        from oauth2client.client import SignedJwtAssertionCredentials

        print('Opening connection to Houseprint sheet')
        # fetch credentials
        json_key = json.load(open(self.gjson))
        scope = ['https://spreadsheets.google.com/feeds']
        credentials = SignedJwtAssertionCredentials(
            json_key['client_email'],
            json_key['private_key'].encode('ascii'),
            scope
        )

        # authorize and login
        gc = gspread.authorize(credentials)
        gc.login()

        # open sheets
        print("Opening spreadsheets")
        sheet = gc.open(self.spreadsheet)
        sites_sheet = sheet.worksheet('Accounts')
        devices_sheet = sheet.worksheet('Devices')
        sensors_sheet = sheet.worksheet('Sensors')

        print('Parsing spreadsheets')
        # 3 sub-methods that parse the different sheets
        self._parse_sites(sites_sheet)
        self._parse_devices(devices_sheet)
        self._parse_sensors(sensors_sheet)

        print('Houseprint parsing complete')

    def _parse_sites(self, sheet):
        """
            Sub method of _parse_sheet() that parses only the 'sites' sheet

            Parameters
            ----------
            sheet: GSpread worksheet
                sheet containing metadata about sites
        """

        records = sheet.get_all_records()

        for r in records:
            if r['Key'] == '':
                continue
            new_site = Site(hp=self,
                            key=r['Key'],
                            size=r['House size'],
                            inhabitants=r['Number of inhabitants'],
                            postcode=r['postcode'],
                            construction_year=r['construction year'],
                            k_level=r['K-level'],
                            e_level=r['E-level'],
                            epc_cert=r['EPC certificate'])
            self.sites.append(new_site)

        print('{} Sites created'.format(len(self.sites)))

    def _parse_devices(self, sheet):
        """
            Sub method of _parse_sheet() that parses only the 'devices' sheet

            Parameters
            ----------
            sheet: GSpread worksheet
                sheet containing metadata about devices
        """

        records = sheet.get_all_records()

        for r in records:
            if r['Key'] == '':
                continue

            # find parent site and check if it exists
            site = self.find_site(r['Parent site'])
            if site is None:
                raise ValueError('Device {} was given an invalid site key {}'.format(r['Key'], r['Parent site']))

            # create a new device according to its manufacturer
            if r['manufacturer'] == 'Flukso':
                new_device = Fluksometer(site=site, key=r['Key'])
            else:
                raise NotImplementedError('Devices from {} are not supported'.format(r['manufacturer']))

            # add new device to parent site
            site.devices.append(new_device)

        print('{} Devices created'.format(sum([len(site.devices) for site in self.sites])))

    def _parse_sensors(self, sheet):
        """
            Sub method of _parse_sheet() that parses only the 'sensors' sheet

            Parameters
            ----------
            sheet: GSpread worksheet
                sheet containing metadata about sensors
        """

        records = sheet.get_all_records()

        for r in records:
            if r['Sensor_id'] == '': continue

            # find parent. If a parent device is specified, us that, otherwise use a parent site directly
            if r['parent device'] != '':
                device = self.find_device(r['parent device'])
                if device is None:
                    raise ValueError(
                        'Sensor {} was given an invalid device key {}. \
                        Leave the device field empty if you want to add a sensor without a device'.format(
                            r['Sensor_id'], r['parent device']))
            else:
                site = self.find_site(r['parent site'])
                if site is None:
                    raise ValueError(
                        'Sensor {} was given an invalid site key {}'.format(r['Sensor_id'], r['parent site']))

            # create new sensor according to its manufacturer
            if r['manufacturer'] == 'Flukso':
                new_sensor = Fluksosensor(
                    device=device,
                    key=r['Sensor_id'],
                    token=r['token'],
                    type=r['sensor type'],
                    description=r['name by user'],
                    system=r['system'],
                    quantity=r['quantity'],
                    unit=r['unit'],
                    direction=r['direction'],
                    tariff=r['tariff'],
                    cumulative=None  # will be determined based on type
                )
            else:
                raise NotImplementedError('Sensors from {} are not supported'.format(r['manufacturer']))

            new_sensor.device.sensors.append(new_sensor)

        print('{} sensors created'.format(sum([len(site.sensors) for site in self.sites])))

    def get_sensors(self, sensortype=None):
        """
            Return a list with all sensors

            Parameters
            ----------
            sensortype: gas, water, electricity: optional

            Returns
            -------
            list of sensors
        """
        res = []
        for site in self.sites:
            for sensor in site.get_sensors(sensortype=sensortype):
                res.append(sensor)
        return res

    def get_fluksosensors(self, **kwargs):
        """
        Same thing as get_sensors, but only for fluksosensors

        Parameters
        ----------
        kwargs

        Returns
        -------
        [Fluksosensor]
        """
        return [sensor for sensor in self.get_sensors(**kwargs) if isinstance(
            sensor, Fluksosensor)]

    def get_devices(self):
        """
            Return a list with all devices

            Returns
            -------
            list of devices
        """
        res = []
        for site in self.sites:
            for device in site.devices:
                res.append(device)
        return res

    def search_sites(self, **kwargs):
        """
            Parameters
            ----------
            kwargs: any keyword argument, like key=mykey

            Returns
            -------
            List of sites satisfying the search criterion or empty list if no
            variable found.
        """

        result = []
        for site in self.sites:
            for keyword, value in kwargs.items():
                if getattr(site, keyword) == value:
                    continue
                else:
                    break
            else:
                result.append(site)

        return result

    def search_sensors(self, **kwargs):
        """
            Parameters
            ----------
            kwargs: any keyword argument, like key=mykey

            Returns
            -------
            List of sensors satisfying the search criterion or empty list if no
            variable found.
        """

        result = []
        for sensor in self.get_sensors():
            for keyword, value in kwargs.items():
                if value in getattr(sensor, keyword):
                    continue
                else:
                    break
            else:
                result.append(sensor)

        return result

    def find_site(self, key):
        """
            Parameters
            ----------
            key: string

            Returns
            -------
            Site
        """

        for site in self.sites:
            if site.key == key:
                return site
        return None

    def find_device(self, key):
        """
            Parameters
            ----------
            key: string

            Returns
            -------
            Device
        """
        for device in self.get_devices():
            if device.key.lower() == key.lower():
                return device
        return None

    def find_sensor(self, key):
        """
            Parameters
            ----------
            key: string

            Returns
            -------
            Sensor
        """
        for sensor in self.get_sensors():
            if sensor.key.lower() == key.lower():
                return sensor
        return None

    def save(self, filename, pickle_format='jsonpickle'):
        """
        Save the houseprint object

        Parameters
        ----------
        * filename : str
            Filename, if relative path or just filename, it is appended to the
            current working directory
        pickle_format : str
            'jsonpickle' or 'pickle'
            pickle may be more robust, but jsonpickle should be compatible
            across python versions

        """
        # temporarily delete tmpo session
        try:
            tmpos_tmp = self._tmpos
            delattr(self, '_tmpos')
        except:
            pass

        abspath = os.path.join(os.getcwd(), filename)

        if pickle_format == 'jsonpickle':
            with open(abspath, 'w') as f:
                frozen = jsonpickle.encode(self)
                f.write(frozen)
        elif pickle_format == 'pickle':
            with open(abspath, 'wb') as f:
                pickle.dump(self, file=f)
        else:
            raise NotImplementedError("Pickle format '{}' is not supported".format(pickle_format))

        print("Saved houseprint to {}".format(abspath))

        # restore tmposession if needed
        try:
            setattr(self, '_tmpos', tmpos_tmp)
        except:
            pass

    def init_tmpo(self, tmpos=None, path_to_tmpo_data=None):
        """
            Flukso sensors need a tmpo session to obtain data.
            It is overkill to have each flukso sensor make its own session, syncing would
            take too long and be overly redundant.
            Passing a tmpo session to the get_data function is also bad form because 
            we might add new types of sensors that don't use tmpo in the future.
            This is why the session is initialised here.

            A tmpo session as parameter is optional.  If passed, no additional sensors are added.
            
            If no session is passed, a new one will be created using the location in the config file.
            It will then be populated with the flukso sensors known to the houseprint object

            Parameters
            ----------

            tmpos : tmpo session
            path_to_tmpo_data : str
        """

        if tmpos is not None:
            self._tmpos = tmpos
        else:
            try:
                path_to_tmpo_data = config.get('tmpo', 'data')
            except:
                path_to_tmpo_data = None

            self._tmpos = tmpo.Session(path_to_tmpo_data)
            self._add_sensors_to_tmpos()

        print("Using tmpo database from {}".format(self._tmpos.db))

    def _add_sensors_to_tmpos(self):
        """
        Add all flukso sensors in the houseprint to the tmpo session
        """

        for sensor in self.get_fluksosensors():
            self._tmpos.add(sensor.key, sensor.token)

    def get_tmpos(self):
        """
            Returns
            -------
            TMPO session
        """
        if hasattr(self, '_tmpos'):
            return self._tmpos
        else:
            self.init_tmpo()
            return self._tmpos

    @property
    def tmpos(self):
        return self.get_tmpos()

    def sync_tmpos(self, http_errors='warn'):
        """
            Add all Flukso sensors to the TMPO session and sync

            Parameters
            ----------
            http_errors : 'raise' | 'warn' | 'ignore'
                default 'warn'
                define what should be done with TMPO Http-errors
        """

        tmpos = self.get_tmpos()
        for sensor in tqdm(self.get_fluksosensors()):
            try:
                warnings.simplefilter('ignore')
                tmpos.sync(sensor.key)
                warnings.simplefilter('default')
            except HTTPError as e:
                warnings.simplefilter('default')
                if http_errors == 'ignore':
                    continue
                elif http_errors == 'warn':
                    warnings.warn(message='Error for SensorID: ' + sensor.key
                    + str(e))
                else:
                    print('Error for SensorID: ' + sensor.key)
                    raise e

    def get_data(self, sensors=None, sensortype=None, head=None, tail=None, diff='default', resample='min',
                 unit='default'):
        """
        Return a Pandas Dataframe with joined data for the given sensors

        Parameters
        ----------
        sensors : list of Sensor objects
            If None, use sensortype to make a selection
        sensortype : string (optional)
            gas, water, electricity. If None, and Sensors = None,
            all available sensors in the houseprint are fetched
        head, tail: timestamps,
        diff : bool or 'default'
            If True, the original data will be differentiated
            If 'default', the sensor will decide: if it has the attribute
            cumulative==True, the data will be differentiated.
        resample : str (default='min')
            Sampling rate, if any.  Use 'raw' if no resampling.
        unit : str , default='default'
            String representation of the target unit, eg m**3/h, kW, ...
        
        """
        if sensors is None:
            sensors = self.get_sensors(sensortype)
        series = [sensor.get_data(head=head, tail=tail, diff=diff, resample=resample, unit=unit) for sensor in sensors]

        # workaround for https://github.com/pandas-dev/pandas/issues/12985
        series = [s for s in series if not s.empty]

        if series:
            df = pd.concat(series, axis=1)
        else:
            df = pd.DataFrame()

        # Add unit as string to each series in the df.  This is not persistent: the attribute unit will get
        # lost when doing operations with df, but at least it can be checked once.
        for s in series:
            try:
                df[s.name].unit = s.unit
            except:
                pass

        return df

    def get_data_dynamic(self, sensors=None, sensortype=None, head=None,
                         tail=None, diff='default', resample='min',
                         unit='default'):
        """
        Yield Pandas Series for the given sensors

        Parameters
        ----------
        sensors : list(Sensor), optional
            If None, use sensortype to make a selection
        sensortype : str, optional
            gas, water, electricity. If None, and Sensors = None,
            all available sensors in the houseprint are fetched
        head : dt.datetime | pd.Timestamp | int, optional
        tail : dt.datetime | pd.Timestamp | int, optional
        diff : bool | str('default')
            If True, the original data will be differentiated
            If 'default', the sensor will decide: if it has the attribute
            cumulative==True, the data will be differentiated.
        resample : str
            default='min'
            Sampling rate, if any.  Use 'raw' if no resampling.
        unit : str
            default='default'
            String representation of the target unit, eg m**3/h, kW, ...

        Yields
        ------
        Pandas.Series
        """
        if sensors is None:
            sensors = self.get_sensors(sensortype)

        for sensor in sensors:
            ts = sensor.get_data(head=head, tail=tail, diff=diff,
                                  resample=resample, unit=unit)
            if ts.empty:
                continue
            else:
                yield ts

    def add_site(self, site):
        """
        Parameters
        ----------
        site : Site
        """
        site.hp = self
        self.sites.append(site)


def load_houseprint_from_file(filename, pickle_format='jsonpickle'):
    """
    Return a static (=anonymous) houseprint object

    Parameters
    ----------
    filename : str
    pickle_format : str
        'jsonpickle' or 'pickle'
        pickle may be more robust, but jsonpickle should be compatible
        across python versions
    """
    if pickle_format == 'jsonpickle':
        with open(filename, 'r') as f:
            hp = jsonpickle.decode(f.read())
    elif pickle_format == 'pickle':
        with open(filename, 'rb') as f:
            hp = pickle.load(file=f)
    else:
        raise NotImplementedError("Pickle format '{}' is not supported".format(pickle_format))

    return hp



# -*- coding: utf-8 -*-
"""
Script for generating a multivariable regression model for a single sensor.
The script will fetch the data, build a model and make graphs.

Created on 26/03/2017 by Roel De Coninck

"""

import sys, os
import matplotlib
matplotlib.use('Agg')

import pandas as pd
from opengrid.library import houseprint, caching, regression, forecastwrapper
from opengrid import config
c = config.Config()

import matplotlib.pyplot as plt
plt.style.use('ggplot')
plt.rcParams['figure.figsize'] = 10,5


def compute(sensorid, start_model, end_model):
    end = pd.Timestamp('now', tz='Europe/Brussels')
    # Create houseprint from saved file, if not available, parse the google spreadsheet
    try:
        hp_filename = os.path.join(c.get('data', 'folder'), 'hp_anonymous.pkl')
        hp = houseprint.load_houseprint_from_file(hp_filename)
        print("Houseprint loaded from {}".format(hp_filename))
    except Exception as e:
        print(e)
        print("Because of this error we try to build the houseprint from source")
        hp = houseprint.Houseprint()
    hp.init_tmpo()

    # Load the cached daily data
    sensor = hp.find_sensor(sensorid)
    cache = caching.Cache(variable='{}_daily_total'.format(sensor.type))
    df_day = cache.get(sensors=[sensor])
    df_day.rename(columns={sensorid: sensor.type}, inplace=True)

    # Load the cached weather data, clean up and compose a combined dataframe
    weather = forecastwrapper.Weather(location=(50.8024, 4.3407), start=start_model, end=end)
    irradiances = [
        (0, 90),  # north vertical
        (90, 90),  # east vertical
        (180, 90),  # south vertical
        (270, 90),  # west vertical
    ]
    orientations = [0, 90, 180, 270]
    weather_data = weather.days(irradiances=irradiances,
                                wind_orients=orientations,
                                heating_base_temperatures=[0, 6, 8, 10, 12, 14, 16, 18]).dropna(axis=1)
    weather_data.drop(['icon', 'summary', 'moonPhase', 'windBearing', 'temperatureMaxTime', 'temperatureMinTime',
                       'apparentTemperatureMaxTime', 'apparentTemperatureMinTime', 'uvIndexTime',
                       'sunsetTime', 'sunriseTime'],
                      axis=1, inplace=True)
    # Add columns for the day-of-week
    for i, d in zip(range(7), ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']):
        weather_data[d] = 0
        weather_data.loc[weather_data.index.weekday == i, d] = 1
    weather_data = weather_data.applymap(float)

    data = pd.concat([df_day, weather_data], axis=1).dropna()
    data = data.tz_convert('Europe/Brussels')

    df = data.resample(rule='MS').sum()
    if len(df) < 2:
        print("Not enough data for building a monthly reference model")
        sys.exit(1)

    # monthly model, statistical validation
    mv = regression.MVLinReg(df.ix[:end_model], sensor.type, p_max=0.03)
    figures = mv.plot(df=df)

    figures[0].savefig(os.path.join(c.get('data', 'folder'), 'figures', 'multivar_model_' + sensorid + '.png'), dpi=100)
    figures[1].savefig(os.path.join(c.get('data', 'folder'), 'figures', 'multivar_results_' + sensorid + '.png'),
                       dpi=100)

    # weekly model, statistical validation
    df = data.resample(rule='W').sum()
    if len(df.ix[:end_model]) < 4:
        print("Not enough data for building a weekly reference model")
        sys.exit(1)
    mv = regression.MVLinReg(df.ix[:end_model], sensor.type, p_max=0.02)
    if len(df.ix[end_model:]) > 0:
        figures = mv.plot(model=False, bar_chart=True, df=df.ix[end_model:])
        figures[0].savefig(
            os.path.join(c.get('data', 'folder'), 'figures', 'multivar_prediction_weekly_' + sensorid + '.png'),
            dpi=100)


if __name__ == '__main__':


    if not len(sys.argv) == 4:
        print("""
        Use of this script: python mreg_sensors.py sensorid from till

        sensorid: (string) sensortoken
        from: (string) starting date for the identification data of the model
        till: (string) end date for the identification data of the model
        """)
        exit(1)


    # parse arguments
    sensorid = sys.argv[1]
    start_model = pd.Timestamp(sys.argv[2], tz='Europe/Brussels')
    end_model = pd.Timestamp(sys.argv[3], tz='Europe/Brussels') #last day of the data period for the model

    compute(sensorid, start_model, end_model)





import unittest
from six import string_types
from pandas.core.frame import DataFrame

from opengrid.library.kmi import *


class KMITest(unittest.TestCase):
    """
    Class for testing the kmi web scraper
    """

    def test_fetch_website(self):
        """
        Check if the URL works
        """
        self.assertIsInstance(fetch_website(), string_types)

    def test_get_kmi_current_month(self):
        """
        Check if the top function returns a dataframe
        """
        self.assertIsInstance(get_kmi_current_month(), DataFrame)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Synchronize the opengrid data to your computer & cache the houseprint

Created on 16/12/2014 by Roel De Coninck
"""

import os
from opengrid.library import houseprint
from opengrid import config

c = config.Config()

hp = houseprint.Houseprint()
print('Sensor data fetched')

filename = os.path.join(c.get('data', 'folder'), 'hp_anonymous.pkl')
hp.save(filename)

hp.init_tmpo()
hp.sync_tmpos()



__author__ = 'Jan Pecinovsky, Roel De Coninck'

"""
A sensor generates a single data stream.
It can have a parent device, but the possibility is also left open for a sensor to stand alone in a site.
It is an abstract class definition which has to be overridden (by eg. a Fluksosensor).

This class contains all metadata concerning the function and type of the sensor (eg. electricity - solar, ...)
"""

from opengrid.library import misc
from opengrid import ureg
import pandas as pd
import tmpo, sqlite3


class Sensor(object):
    def __init__(self, key=None, device=None, site=None, type=None,
                 description=None, system=None, quantity=None, unit=None,
                 direction=None, tariff=None, cumulative=None):
        self.key = key
        self.device = device
        self.site = site
        self.type = type
        self.description = description
        self.system = system
        self.quantity = quantity
        self.unit = unit
        self.direction = direction
        self.tariff = tariff
        self.cumulative = cumulative

    def __repr__(self):
        return """
    {}
    Key: {}
    Type: {}
    """.format(self.__class__.__name__,
               self.key,
               self.type
               )

    def get_data(self, head=None, tail=None, resample='min'):
        """
        Return a Pandas Series with measurement data

        Parameters
        ----------
        head, tail: timestamps for the begin and end of the interval

        Notes
        -----
        This is an abstract method, because each type of sensor has a different way of fetching the data.

        Returns
        -------
        Pandas Series
        """

        raise NotImplementedError("Subclass must implement abstract method")

    def _get_default_unit(self, diff=True, resample='min'):
        """
        Return a string representation of the default unit for the requested operation
        If there is no unit, returns None

        Parameters
        ----------
        diff : True (default) or False
            If True, the original data has been differentiated
        resample : str (default='min')
            Sampling rate, if any.  Use 'raw' if no resampling.

        Returns
        -------
        target : str or None
            String representation of the target unit, eg m3/h, kW, ...
        """

        if self.type in ['electricity', 'gas', 'heat', 'energy']:
            if diff:
                target = 'W'
            else:
                target = 'kWh'
        elif self.type == 'water':
            if diff:
                target = 'l/min'
            else:
                target = 'liter'
        elif self.type == 'temperature':
            target = 'degC'
        elif self.type == 'pressure':
            target = 'Pa'
        elif self.type in ['battery']:
            target = 'V'
        elif self.type in ['current']:
            target = 'A'
        elif self.type in ['light']:
            target = 'lux'
        elif self.type == 'humidity':
            target = 'percent'
        elif self.type in ['error', 'vibration', 'proximity']:
            target = ''

        else:
            target = None

        return target

    def _unit_conversion_factor(self, diff=True, resample='min', target='default'):
        """
        Return a conversion factor to convert the obtained data
        The method starts from the unit of the sensor, and takes
        into account sampling, differentiation (if any) and target unit.

        For gas, a default calorific value of 10 kWh/liter is used.

        For some units, unit conversion does not apply, and 1.0 is returned.

        Parameters
        ----------
        diff : True (default) or False
            If True, the original data has been differentiated
        resample : str (default='min')
            Sampling rate, if any.  Use 'raw' if no resampling.
        target : str , default='default'
            String representation of the target unit, eg m3/h, kW, ...
            If None, 1.0 is returned

        Returns
        -------
        cf : float
            Multiplication factor for the original data to the target unit
        """

        # get the target
        if target == 'default':
            target = self._get_default_unit(diff=diff, resample=resample)
        if target is None:
            return 1.0

        if resample == 'raw':
            if diff:
                raise NotImplementedError("Differentiation always needs a sampled dataframe")

        # get the source
        if not self.type == 'gas':
            if not diff:
                source = self.unit
            else:
                # differentiation. Careful, this is a hack of the unit system.
                # we have to take care manually of some corner cases
                if self.unit:
                    source = self.unit + '/' + resample
                else:
                    source = self.unit

            return misc.unit_conversion_factor(source, target)
        else:
            # for gas, we need to take into account the calorific value
            # as of now, we use 10 kWh/l by default
            CALORIFICVALUE = 10
            q_src = 1 * ureg(self.unit)
            q_int = q_src * ureg('Wh/liter')
            if not diff:
                source = str(q_int.units)  # string representing the unit, mostly kWh
            else:
                source = str(q_int.units) + '/' + resample
            return CALORIFICVALUE * misc.unit_conversion_factor(source, target)

    def last_timestamp(self, epoch=False):
        """
        Get the last timestamp for a sensor

        Parameters
        ----------
        epoch : bool
            default False
            If True return as epoch
            If False return as pd.Timestamp

        Returns
        -------
        pd.Timestamp | int
        """
        raise NotImplementedError("Subclass must implement abstract method")


class Fluksosensor(Sensor):
    def __init__(self, key=None, token=None, device=None, type=None,
                 description=None, system=None, quantity=None, unit=None,
                 direction=None, tariff=None, cumulative=None, tmpos=None):

        # invoke init method of abstract Sensor
        super(Fluksosensor, self).__init__(key=key,
                                           device=device,
                                           site=device.site if device else None,
                                           type=type,
                                           description=description,
                                           system=system,
                                           quantity=quantity,
                                           unit=unit,
                                           direction=direction,
                                           tariff=tariff,
                                           cumulative=cumulative)

        if token != '':
            self.token = token
        else:
            self.token = device.mastertoken

        if self.unit == '' or self.unit is None:
            if self.type in ['water', 'gas']:
                self.unit = 'liter'
            elif self.type == 'electricity':
                self.unit = 'Wh'
            elif self.type == 'pressure':
                self.unit = 'Pa'
            elif self.type == 'temperature':
                self.unit = 'degC'
            elif self.type == 'battery':
                self.unit = 'V'
            elif self.type == 'light':
                self.unit = 'lux'
            elif self.type == 'humidity':
                self.unit = 'percent'
            elif self.type in ['error', 'vibration', 'proximity']:
                self.unit = ''


        if self.cumulative == '' or self.cumulative is None:
            if self.type in ['water', 'gas', 'electricity', 'vibration']:
                self.cumulative = True
            else:
                self.cumulative = False

        self._tmpos = tmpos

    @property
    def tmpos(self):
        if self._tmpos is not None:
            return self._tmpos
        elif self.device is not None:
            return self.device.tmpos
        else:
            raise AttributeError('TMPO session not defined')

    @property
    def has_data(self):
        """
        Checks if a sensor actually has data by checking the length of the
        tmpo block list

        Returns
        -------
        bool
        """
        tmpos = self.site.hp.get_tmpos()
        return len(tmpos.list(self.key)[0]) != 0

    def get_data(self, head=None, tail=None, diff='default', resample='min', unit='default', tz='UTC'):
        """
        Connect to tmpo and fetch a data series

        Parameters
        ----------
        sensors : list of Sensor objects
            If None, use sensortype to make a selection
        sensortype : string (optional)
            gas, water, electricity. If None, and Sensors = None,
            all available sensors in the houseprint are fetched
        head, tail: timestamps
            Can be epoch, datetime of pd.Timestamp, with our without timezone (default=UTC)
        diff : bool or 'default'
            If True, the original data will be differentiated
            If 'default', the sensor will decide: if it has the attribute
            cumulative==True, the data will be differentiated.
        resample : str (default='min')
            Sampling rate, if any.  Use 'raw' if no resampling.
        unit : str , default='default'
            String representation of the target unit, eg m**3/h, kW, ...
        tz : str, default='UTC'
            Specify the timezone for the index of the returned dataframe

        Returns
        -------
        Pandas Series with additional attribute 'unit' set to
        the string representation of the unit of the data.
        """

        if head is None:
            head = 0
        if tail is None:
            tail = 2147483647  # tmpo epochs max

        data = self.tmpos.series(sid=self.key, head=head, tail=tail)

        if data.dropna().empty:
            # Return an empty dataframe with correct name
            return pd.Series(name=self.key)

        data = data.tz_convert(tz)

        if resample != 'raw':

            if resample == 'hour':
                rule = 'H'
            elif resample == 'day':
                rule = 'D'
            else:
                rule = resample

            # interpolate to requested frequency
            newindex = data.resample(rule).first().index
            data = data.reindex(data.index.union(newindex))
            data = data.interpolate(method='time')
            data = data.reindex(newindex)

            if diff == 'default':
                diff = self.cumulative

            if diff:
                data = data.diff()

        # unit conversion
        if unit == 'default':
            unit = self._get_default_unit(diff=diff, resample=resample)
        ucf = self._unit_conversion_factor(diff=diff, resample=resample, target=unit)
        data *= ucf
        data.unit = unit

        return data

    def last_timestamp(self, epoch=False):
        """
        VERY HACKY CODE, WAITING FOR A PULL-REQUEST IN TMPO TO GO THROUGH

            Get the theoretical last timestamp for a sensor
            It is the mathematical end of the last block, the actual last sensor stamp may be earlier

            Parameters
            ----------
            epoch : bool
                default False
                If True return as epoch
                If False return as pd.Timestamp

            Returns
            -------
            pd.Timestamp | int
        """
        query = tmpo.SQL_TMPO_LAST
        sid = self.key
        tmpos = self.site.hp.get_tmpos()

        dbcon = sqlite3.connect(tmpos.db)
        dbcur = dbcon.cursor()
        dbcur.execute(tmpo.SQL_SENSOR_TABLE)
        dbcur.execute(tmpo.SQL_TMPO_TABLE)

        query = dbcur.execute(query, (sid,))
        try:
            rid, lvl, bid = query.fetchone()
        except TypeError:
            return None

        end_of_block = tmpos._blocktail(lvl, bid)

        dbcon.close()

        if epoch:
            return end_of_block
        else:
            return pd.Timestamp.fromtimestamp(end_of_block).tz_localize('UTC')


# -*- coding: utf-8 -*-
"""
Created on Thu Jan  7 09:36:16 2016

@author: roel
"""

# -*- coding: utf-8 -*-
"""
Created on Mon Dec 30 02:37:25 2013

@author: roel
"""

import os, sys
import unittest
import inspect
import numpy as np
import pandas as pd
import datetime as dt

test_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(test_dir)
# add the path to opengrid to sys.path
sys.path.insert(1, os.path.join(test_dir, os.pardir, os.pardir, os.pardir))
from opengrid.library import analysis

# Note: there is a opengrid.cfg in the test_dir which is loaded here!!
from opengrid import config
cfg = config.Config()

class AnalysisTest(unittest.TestCase):
    


    def test_init_DailyAgg(self):
        "Initialize a DailyAgg, check attributes of result"
        index = pd.DatetimeIndex(start='20160101 00:51:15', freq='h', periods=80)
        df = pd.DataFrame(index=index, data=np.arange(0, 80, 1), columns=['A'])

        anls = analysis.DailyAgg(df, agg='min')
        self.assertTrue(hasattr(anls, 'result'))

    def test_DailyAgg_analysis(self):
        "Verify result of analyses"
        index = pd.DatetimeIndex(start='20160101 00:51:15', freq='h', periods=80)
        df = pd.DataFrame(index=index, data={'A':np.arange(0, 80, 1), 'B':np.zeros(80)})

        anls = analysis.DailyAgg(df, agg='min')
        dayindex = pd.DatetimeIndex(start='20160101', freq='D', periods=4)
        self.assertTrue((anls.result.index==dayindex).all())
        self.assertTrue((anls.result['A'] == pd.Series(data=[0,24,48,72], index=dayindex)).all())

    def test_DailyAgg_analysis_with_timelimits(self):
        "Check results when limiting the hours for the analysis"
        index = pd.DatetimeIndex(start='20160101 00:51:15', freq='h', periods=80)
        df = pd.DataFrame(index=index, data={'A':np.arange(0, 80, 1), 'B':np.zeros(80)})

        anls = analysis.DailyAgg(df, agg='min', starttime=dt.time(hour=3, minute=34), endtime=dt.time(hour=22, minute=55))
        dayindex = pd.DatetimeIndex(start='20160101', freq='D', periods=4)
        self.assertTrue((anls.result.index==dayindex).all())
        self.assertTrue((anls.result['A'] == pd.Series(data=[3,27,51,75], index=dayindex)).all())

    def test_json(self):
        "Verify the to_json method"
        index = pd.DatetimeIndex(start='20160101 00:51:15', freq='h', periods=80)
        df = pd.DataFrame(index=index, data={'A':np.arange(0, 80, 1), 'B':np.zeros(80)})

        js = analysis.DailyAgg(df, agg='min', starttime=dt.time(hour=3, minute=34), endtime=dt.time(hour=22, minute=55)).to_json()
        self.assertIsInstance(pd.read_json(js), pd.DataFrame)

    def test_analysis_with_empty_df(self):
        "Empty dataframes should not raise errors"

        anls = analysis.DailyAgg(df=pd.DataFrame(), agg='max')
        self.assertIsInstance(anls.result, pd.DataFrame)

        js = anls.to_json()
        self.assertIsInstance(pd.read_json(js), pd.DataFrame)

    def test_run_analysis_twice(self):
        "Running an analysis twice should give same result"
        index = pd.DatetimeIndex(start='20160101 00:51:15', freq='h', periods=80)
        df = pd.DataFrame(index=index, data={'A':np.arange(0, 80, 1), 'B':np.zeros(80)})

        anls = analysis.DailyAgg(df, agg='min', starttime=dt.time(hour=3, minute=34), endtime=dt.time(hour=22, minute=55))
        result1 = anls.result.copy()
        anls.do_analysis(agg='min', starttime=dt.time(hour=3, minute=34), endtime=dt.time(hour=22, minute=55))
        result2 = anls.result.copy()
        self.assertTrue((result1==result2).all().all())

    def test_run_analysis_twice_different_args(self):
        "Running an analysis twice should give same result"
        index = pd.DatetimeIndex(start='20160101 00:51:15', freq='h', periods=80)
        df = pd.DataFrame(index=index, data={'A':np.arange(0, 80, 1), 'B':np.zeros(80)})

        anls = analysis.DailyAgg(df, agg='min', starttime=dt.time(hour=3, minute=34), endtime=dt.time(hour=22, minute=55))
        result1 = anls.result.copy()
        anls.do_analysis(agg='max', starttime=dt.time(hour=3, minute=34), endtime=dt.time(hour=22, minute=55))
        result2 = anls.result.copy()
        self.assertFalse((result1==result2).all().all())



if __name__ == '__main__':
    
    unittest.main()



# coding: utf-8

# ### This script shows how to use the existing code in opengrid to create an interactive timeseries plot.

# In[ ]:

import os
import pytz
BXL = pytz.timezone('Europe/Brussels')
import pandas as pd
import charts
from time import time

from opengrid import config
c=config.Config()
DEV = c.get('env', 'type') == 'dev' # DEV is True if we are in development environment, False if on the droplet

from opengrid.library import houseprint


# ### Script settings
# 

# In[ ]:

number_of_days = 7


# ### We create a dataframe per sensor

# In[ ]:

# Load houseprint from cache if possible, otherwise build it from source
try:
    hp_filename = os.path.join(c.get('data', 'folder'), 'hp_anonymous.pkl')
    hp = houseprint.load_houseprint_from_file(hp_filename)
    print("Houseprint loaded from {}".format(hp_filename))
except Exception as e:
    print(e)
    print("Because of this error we try to build the houseprint from source")
    hp = houseprint.Houseprint()


# In[ ]:

start = pd.Timestamp(time() - number_of_days*86400, unit='s')


# In[ ]:

sensors = hp.get_sensors()
#sensors.remove('b325dbc1a0d62c99a50609e919b9ea06')


# In[ ]:

for sensor in sensors:
    s = sensor.get_data(head=start)
    # do an ugly operation to get localtime.  A simple tz_convert does not work: charts will always plot 
    # in UTC.
    new_index_list = []
    for index_utc in s.index:
        new_index_list.append(index_utc + index_utc.tz_convert(BXL).utcoffset())
    s.index = pd.DatetimeIndex(new_index_list)
    try:    
        # plot with charts (don't show it) and save html
        charts.plot(pd.DataFrame(s), stock=True, 
                    save=os.path.join(c.get('data', 'folder'), 'figures', 'TimeSeries_'+sensor.key+'.html'), show=False)
    except Exception as e:
        print(e)


# In[ ]:






from opengrid.library import analysis
import logging
from scipy import stats
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as fm
from statsmodels.sandbox.regression.predstd import wls_prediction_std
from copy import deepcopy


class MVLinReg(analysis.Analysis):
    """
    Multi-variable linear regression based on statsmodels and Ordinary Least Squares (ols)

    Pass a dataframe with the variable to be modelled (endogenous variable) and the possible independent (exogenous)
    variables.  Specify as string the name of the endogenous variable, and optionally pass a list with names of
    exogenous variables to try (by default all other columns will be tried as exogenous variables).

    The analysis is based on a forward-selection approach: starting from a simple model, the model is iteratively
    refined and verified until no statistical relevant improvements can be obtained.  Each model in the iteration loop
    is stored in the attribute self.list_of_fits.  The selected model is self.fit (=pointer to the last element of
    self.list_of_fits).

    The dataframe can contain daily, weekly, monthly, yearly ... values.  Each row is an instance.


    Examples
    --------

    >> mvlr = MVLinReg(df, 'gas', p_max=0.04)
    >> mvlr = MVLinReg(df, 'gas', list_of_exog=['heatingDegreeDays14', 'GlobalHorizontalIrradiance', 'WindSpeed'])


    """

    def __init__(self, df, endog, **kwargs):
        """

        Parameters
        ----------
        df : pd.DataFrame
            Datetimeindex and both endogenous and exogenous variables as columns
        endog : str
            Name of the endogeneous variable to model
        p_max : float (default=0.05)
            Acceptable p-value of the t-statistic for estimated parameters
        list_of_exog : list of str (default=None)
            If None (default), try to build a model with all columns in the dataframe
            If a list with column names is given, only try these columns as exogenous variables
        confint : float, default=0.05
            Two-sided confidence interval for predictions.
        cross_validation : bool, default=False
            If True, compute the model based on cross-validation (leave one out)
            Only possible if the df has less than 15 entries.
            Note : this will take much longer computation times!
        allow_negative_predictions : bool, default=False
            If True, allow predictions to be negative.
            For gas consumption or PV production, this is not physical so allow_negative_predictions should be False
        """
        self.df = df.copy()
        assert endog in self.df.columns, "The endogenous variable {} is not a column in the dataframe".format(endog)
        self.endog = endog

        self.p_max = kwargs.get('p_max', 0.05)
        self.list_of_exog = kwargs.get('list_of_exog', self.df.columns.tolist())
        self.confint = kwargs.get('confint', 0.05)
        self.cross_validation = kwargs.get('cross_validation', False)
        self.allow_negative_predictions = kwargs.get('allow_negative_predictions', False)
        try:
            self.list_of_exog.remove(self.endog)
        except:
            pass

        self.do_analysis()


    def do_analysis(self):
        """
        Find the best model (fit) and create self.list_of_fits and self.fit

        """
        if self.cross_validation:
            return self._do_analysis_cross_validation()
        else:
            return self._do_analysis_no_cross_validation()


    def _do_analysis_no_cross_validation(self):
        """
        Find the best model (fit) and create self.list_of_fits and self.fit

        """

        self.list_of_fits = []
        # first model is just the mean
        self.list_of_fits.append(fm.ols(formula='{} ~ 1'.format(self.endog), data=self.df).fit())
        # try to improve the model until no improvements can be found
        all_exog = self.list_of_exog[:]
        while all_exog:
            # try each x in all_exog and overwrite the best_fit if we find a better one
            # the first best_fit is the one from the previous round
            best_fit = deepcopy(self.list_of_fits[-1])
            for x in all_exog:
                # make new_fit, compare with best found so far
                formula = self.list_of_fits[-1].model.formula + '+{}'.format(x)
                fit = fm.ols(formula=formula, data=self.df).fit()
                best_fit = self.find_best_bic([best_fit, fit])

            # Sometimes, the obtained fit may be better, but contains unsignificant parameters.
            # Correct the fit by removing the unsignificant parameters and estimate again
            best_fit = self._prune(best_fit, p_max=self.p_max)

            # if best_fit does not contain more variables than last fit in self.list_of_fits, exit
            if best_fit.model.formula in self.list_of_fits[-1].model.formula:
                break
            else:
                self.list_of_fits.append(best_fit)
                all_exog.remove(x)
        self.fit = self.list_of_fits[-1]


    def _do_analysis_cross_validation(self):
        """
        Find the best model (fit) based on cross-valiation (leave one out)

        """
        assert len(self.df) < 15, "Cross-validation is not implemented if your sample contains more than 15 datapoints"

        # initialization: first model is the mean, but compute cv correctly.
        errors = []
        formula = '{} ~ 1'.format(self.endog)
        for i in self.df.index:
            # make new_fit, compute cross-validation and store error
            df_ = self.df.drop(i, axis=0)
            fit = fm.ols(formula=formula, data=df_).fit()
            cross_prediction = self._predict(fit=fit, df=self.df.loc[[i], :])
            errors.append(cross_prediction['predicted'] - cross_prediction[self.endog])

        self.list_of_fits = [fm.ols(formula=formula, data=self.df).fit()]
        self.list_of_cverrors = [np.mean(np.abs(np.array(errors)))]

        # try to improve the model until no improvements can be found
        all_exog = self.list_of_exog[:]
        while all_exog:
            #import pdb;pdb.set_trace()
            # try each x in all_exog and overwrite if we find a better one
            # at the end of iteration (and not earlier), save the best of the iteration
            better_model_found = False
            best = dict(fit=self.list_of_fits[-1], cverror=self.list_of_cverrors[-1])
            for x in all_exog:
                formula = self.list_of_fits[-1].model.formula + '+{}'.format(x)
                # cross_validation, currently only implemented for monthly data
                # compute the mean error for a given formula based on leave-one-out.
                errors = []
                for i in self.df.index:
                    # make new_fit, compute cross-validation and store error
                    df_ = self.df.drop(i, axis=0)
                    fit = fm.ols(formula=formula, data=df_).fit()
                    cross_prediction = self._predict(fit=fit, df=self.df.loc[[i], :])
                    errors.append(cross_prediction['predicted'] - cross_prediction[self.endog])
                cverror = np.mean(np.abs(np.array(errors)))
                # compare the model with the current fit
                if  cverror < best['cverror']:
                    # better model, keep it
                    # first, reidentify using all the datapoints
                    best['fit'] = fm.ols(formula=formula, data=self.df).fit()
                    best['cverror'] = cverror
                    better_model_found = True

            if better_model_found:
                self.list_of_fits.append(best['fit'])
                self.list_of_cverrors.append(best['cverror'])
            else:
                # if we did not find a better model, exit
                break

            # next iteration with the found exog removed
            all_exog.remove(x)

        self.fit = self.list_of_fits[-1]




    def _prune(self, fit, p_max):
        """
        If the fit contains statistically insignificant parameters, remove them.
        Returns a pruned fit where all parameters have p-values of the t-statistic below p_max

        Parameters
        ----------
        fit: fm.ols fit object
            Can contain insignificant parameters
        p_max : float
            Maximum allowed probability of the t-statistic

        Returns
        -------
        fit: fm.ols fit object
            Won't contain any insignificant parameters

        """

        for par in fit.pvalues.where(fit.pvalues > p_max).dropna().index:

            corrected_formula = fit.model.formula.replace('+{}'.format(par), '')
            fit = fm.ols(formula=corrected_formula, data=self.df).fit()
        return fit

    def find_best_rsquared(self, list_of_fits):
        """Return the best fit, based on rsquared"""
        res = sorted(list_of_fits, key=lambda x: x.rsquared)
        return res[-1]

    def find_best_akaike(self, list_of_fits):
        """Return the best fit, based on Akaike information criterion"""
        res = sorted(list_of_fits, key=lambda x: x.aic)
        return res[0]

    def find_best_bic(self, list_of_fits):
        """Return the best fit, based on Akaike information criterion"""
        res = sorted(list_of_fits, key=lambda x: x.bic)
        return res[0]


    def _predict(self, fit, df, **kwargs):
        """
        Return a df with predictions and confidence interval
        The df will contain the following columns:
        - 'predicted': the model output
        - 'interval_u', 'interval_l': upper and lower confidence bounds.

        Parameters
        ----------
        fit : Statsmodels fit
        df : pandas DataFrame or None (default)
            If None, use self.df
        confint : float (default=0.05)
            Confidence level for two-sided hypothesis, if given, overrides the default one.

        Returns
        -------
        df : pandas DataFrame
            same as df with additional columns 'predicted', 'interval_u' and 'interval_l'
        """


        confint = kwargs.get('confint', self.confint)

        # Add model results to data as column 'predictions'
        if 'Intercept' in fit.model.exog_names:
            df['Intercept'] = 1.0
        df['predicted'] = fit.predict(df)
        if not self.allow_negative_predictions:
            df.loc[df['predicted'] < 0, 'predicted'] = 0
        prstd, interval_l, interval_u = wls_prediction_std(fit, df[fit.model.exog_names])
        df['interval_l'] = interval_l
        df['interval_u'] = interval_u

        return df

    def predict(self, **kwargs):
        """
        Add predictions and confidence interval to self.df
        self.df will contain the following columns:
        - 'predicted': the model output
        - 'interval_u', 'interval_l': upper and lower confidence bounds.

        Parameters
        ----------
        confint : float (default=0.05)
            Confidence level for two-sided hypothesis, if given, overrides the default one.

        Returns
        -------
        Nothing, adds columns to self.df
        """
        self.df = self._predict(fit=self.fit, df=self.df, **kwargs)

    def plot(self, model=True, bar_chart=True, **kwargs):
        """
        Plot measurements and predictions.

        By default, use self.fit and self.df, but both can be overruled by the arguments df and fit
        This function will detect if the data has been used for the modelling or not and will
        visualize them differently.

        Parameters
        ----------
        model : boolean, default=True
            If True, show the modified energy signature
        bar_chart : boolean, default=True
            If True, make a bar chart with predicted and measured data
        df : pandas Dataframe, default=None
            The data to be plotted.  If None, use self.df
            If the dataframe does not have a column 'predicted', a prediction will be made
        fit : statsmodels fit, default=None
            The model to be used.  if None, use self.fit
        confint : float (default=0.05)
            Confidence level for two-sided hypothesis, if given, overrides the default one.

        Returns
        -------
        figures : List of plt.figure objects.

        """
        figures = []
        fit = kwargs.get('fit', self.fit)
        df = kwargs.get('df', self.df)

        if not 'predicted' in df.columns:
            df = self._predict(fit=fit, df=df, confint=kwargs.get('confint', 0.05))
        # split the df in the auto-validation and prognosis part
        try:
            df_auto = df.loc[self.df.index, :]
        except KeyError:
            # no overlap between self.df and df
            df_auto = pd.DataFrame()
            df_prog = df
        else:
            df_prog = df.drop(self.df.index)
            if len(df_prog) > 0:
                assert df_prog.index[0] > df_auto.index[-1], "Back-casting is currently not implemented"

        if model:
            # The first variable in the formula is the most significant.  Use it as abcis for the plot
            try:
                exog1 = fit.model.formula.split('+')[1].strip()
            except IndexError:
                exog1 = self.list_of_exog[0]

            # plot model as an adjusted trendline
            # get sorted model values
            dfmodel = df[[exog1, 'predicted', 'interval_u', 'interval_l']]
            dfmodel.index = dfmodel[exog1]
            dfmodel.sort(inplace=True)
            plt.plot(dfmodel.index, dfmodel['predicted'], '--', color='royalblue')
            plt.plot(dfmodel.index, dfmodel['interval_l'], ':',  color='royalblue')
            plt.plot(dfmodel.index, dfmodel['interval_u'], ':', color='royalblue')
            # plot dots for the measurements
            if len(df_auto) > 0:
                plt.plot(df_auto[exog1], df_auto[self.endog], 'o', mfc='orangered', mec='orangered', ms=8, label='Data used for model fitting')
            if len(df_prog) > 0:
                plt.plot(df_prog[exog1], df_prog[self.endog], 'o', mfc='seagreen', mec='seagreen', ms=8, label='Data not used for model fitting')
            plt.title('{} - rsquared={} - BIC={}'.format(fit.model.formula, fit.rsquared, fit.bic))
            figures.append(plt.gcf())

        if bar_chart:
            ind = np.arange(len(df.index))  # the x locations for the groups
            width = 0.35  # the width of the bars

            fig, ax = plt.subplots()
            title = 'Measured' # will be appended based on the available data
            if len(df_auto) > 0:
                model = ax.bar(ind[:len(df_auto)], df_auto['predicted'], width*2, color='#FDD787', ecolor='#FDD787',
                               yerr=df_auto['interval_u'] - df_auto['predicted'], label=self.endog + ' modelled')
                title = title + ', modelled'
            if len(df_prog) > 0:
                prog = ax.bar(ind[len(df_auto):], df_prog['predicted'], width * 2, color='#6CD5A1', ecolor='#6CD5A1',
                              yerr=df_prog['interval_u'] - df_prog['predicted'], label=self.endog + ' expected')
                title = title + ' and predicted'

            meas = ax.bar(ind+width/2., df[self.endog], width, label=self.endog + ' measured', color='#D5756C')
            # add some text for labels, title and axes ticks
            ax.set_ylabel(self.endog)
            ax.set_title('{} {}'.format(title, self.endog))
            ax.set_xticks(ind + width)
            ax.set_xticklabels([x.strftime('%d-%m-%Y') for x in df.index], rotation='vertical')
            ax.yaxis.grid(True)
            ax.xaxis.grid(False)


            plt.legend(ncol=3, loc='upper center')
            figures.append(plt.gcf())

        plt.show()

        return figures

class LinearRegression(analysis.Analysis):
    """
    Calculate a simple linear regression given a dataframe with X and Y values
    """

    def __init__(self, independent, dependent, *args, **kwargs):
        """
        Parameters
        ----------
        independent : pandas.Series
        dependent : pandas.Series
        """
        df = pd.concat([independent, dependent], axis=1).dropna()
        df.columns = ['independent', 'dependent']
        self._check_df(df=df)
        super(LinearRegression, self).__init__(df=df, *args, **kwargs)

    def do_analysis(self, *args, **kwargs):
        df = self._calculate_regression_data()
        slope, self.intercept, r_value, self.p_value, self.std_err = stats.linregress(
            df.independent, df.dependent)
        self._set_slope(slope)
        self._set_r_value(r_value)
        #self.rsquared = self.r_value**2
        try:
            self.rsquared = self._r2()
        except AttributeError:
            # the subclass that gets this error will have to find another way to calculate rsquared
            pass

    def _check_df(self, df):
        """
        Check if dataframe has correct size etc.
        Throws error if dataframe is not good.
        """
        if not len(df) > 2:
            raise ValueError("Data set is too small")

    def _r2(self):
        """
            Calculate R Squared

            Returns
            -------
            float
        """
        res = np.corrcoef(self.df.dependent, self.get_y(self.df.independent))[0, 1] ** 2
        return res

    def _calculate_regression_data(self):
        """
            Decide what data to use for the linear regression.
            In this case all data.
        """
        return self.df

    def _set_slope(self, slope):
        """
        Sets slope attribute, but throws an error when it is negative

        Parameters
        ----------
        slope: float

        """
        self.slope = slope
        if slope < 0:
            raise ValueError("Negative slope!")

    def _set_r_value(self, r_value):
        """
        Sets r_value attribute, but throws a warning when it is negative

        Parameters
        ----------
        r_value: float
        """
        self.r_value = r_value
        if r_value < 0:
            logging.warning("Negative r_value")

    def get_y(self, x):
        """
        Calculate the value on the trend line for a given x-value or an array of x-values

        Parameters
        ----------
        x: single number (float or int)
            OR iterable (array) of numbers

        Returns
        -------
        float if x is a single number
        array of floats if x is an iterable
        """

        # check if x is an iterable
        if not hasattr(x, '__iter__'):
            return self._get_y(x)
        else:
            return [self._get_y(val) for val in x]

    def _get_y(self, x):
        """
        Calculate a value on the trend line for a given x-value

        Parameters
        ----------
        x: number

        Returns
        -------
        float
        """

        # y = ax + b
        return self.slope * x + self.intercept

    def get_x(self, y):
        """
        Calculate value on the trendline for given y-value

        Parameters
        ----------
        y: number

        Return
        ------
        float
        """

        # y = ax + b => x = (y-b)/a
        return (y - self.intercept) / self.slope

    def plot(self, force_origin=True):
        """
        Create scatterplot and plot trendline

        Parameters
        ----------
        force_origin: boolean, default True
            set the default values of the axis to 0,0

        Returns
        -------
        matplotlib figure
        """
        fig = plt.figure()
        ax1 = fig.add_subplot(111)

        ax1.scatter(self.df.independent, self.df.dependent, alpha=1, s=20)

        function = "${s:.2f}x + {i:.2f}$".format(s=self.slope, i=self.intercept)
        r2 = "$R^2$: ${r:.2f}$".format(r=self.rsquared)
        label = "{}\n{}".format(function, r2)

        x = self.get_trend_x()

        ax1.plot(x, self.get_y(x), '-', label=label)
        plt.legend(loc='upper left')

        if force_origin:
            x1, x2, y1, y2 = plt.axis()
            plt.axis((0, x2, 0, y2))

        return fig

    def get_trend_x(self):
        """
        Return the values that are needed to draw the trendline.
        In this case, the min and max since it's a single line
        """

        ret = [self.df.independent.min(), self.df.independent.max()]
        return ret

    def score_spacing(self, max_spacing, min_spacing=0.0, spacing_tolerance=0.1, **kwargs):
        """
        Assign a score between 0 and 1 to the spacing of the x-values.
        We need a nominal spacing, between a miniumum and maximum.
        The spacing of this regression will be given as a percentage of that nominal.

        A spacing_tolerance can be set to allow values that lie close to the nominal spacing to get a value of 1.

        Parameters
        ----------
        max_spacing : float
        min_spacing : float (default 0)
        spacing_tolerance : float (default 0.1)
            percentage: only values between 0 and 1 allowed

        Returns
        -------
        float (percentage between 0 and 1)

        """
        margin = 1 - spacing_tolerance  # eg. if the spacing_tolerance is 5%, we will lower the nominal to 95% of it's value

        diff_x = self.df.independent.max() - self.df.independent.min()
        nominal_diff_x = (max_spacing - min_spacing) * margin

        return min(1, float(diff_x) / float(nominal_diff_x))

    def score_observations(self, expected_observations, **kwargs):
        """
        Assign a score between 0 and 1 to the number of observations.

        eg. If the analysis runs on monthly values for a year, the expected_observations number of observations is 12.
        If the dataframe only has 6 observations, the score will be 0.5

        Parameters
        ----------
        expected_observations: int

        Returns
        -------
        float (between 0 and 1)
        """

        return min(1, float(len(self.df)) / float(expected_observations))

    def score_total(self, weight=0.5, **kwargs):
        """
        Assign a score between 0 and 1 to the quality of the regression data as a whole

        This is done by multiplying the R Squared value with the the other scores.
        The other scores are weighted in to limit their influence on the final score.

        Parameters
        ----------
        weight: float

        Returns
        -------
        float (between 0 and 1)
        """
        spacing = self.score_spacing(**kwargs)
        observations = self.score_observations(**kwargs)

        return (self.rsquared * self._weigh_score(spacing, weight) * self._weigh_score(observations, weight))

    @staticmethod
    def _weigh_score(score, weight):
        """
        Adjusts a score to a weight.

        eg. If a score is 0.5 but you only want it to have an impact of 30% (weight = 0.3),
        the score will be redistributed between 0.7 and 1, resulting to 0.85.

        Parameters
        ----------
        score: float
        weight: float

        Returns
        -------
        float
        """
        return 1 - weight + (score * weight)


class LinearRegression2(LinearRegression):
    """
        Calculate a linear regression that uses a predefined breakpoint to break between regression and baseload
    """

    def __init__(self, independent, dependent, breakpoint, *args, **kwargs):
        """
        Parameters
        ----------
        independent : pandas.Series
        dependent : pandas.Series
        breakpoint : int | float
            value on the x-axis where to break between regression and baseload
        """

        self.breakpoint = breakpoint
        super(LinearRegression2, self).__init__(independent=independent,
                                                dependent=dependent, *args,
                                                **kwargs)

    def do_analysis(self, *args, **kwargs):
        # base_load is the mean of all y values where x is below the breakpoint
        self._set_base_load(self.df[self.df.independent <= self.breakpoint].dependent.mean())
        super(LinearRegression2, self).do_analysis(*args, **kwargs)

        if self.base_load is not None:
            self._set_intersect(self.get_x(self.base_load))
        else:
            self._set_intersect(None)

        self.rsquared = self._r2()

    def _set_intersect(self, intersect):
        """
        Sets intersect attribute, checks for validity
        It has to be positive

        Parameters
        ----------
        intersect: float

        """
        if intersect is not None and intersect > 0:
            self.intersect = intersect
        else:
            self.intersect = None

    def _set_slope(self, slope):
        """
        Sets slope attribute, throws an error if it is negative

        Parameters
        ----------
        slope: float
        """
        if slope < 0:
            raise ValueError('Negative slope not possible for this type of regression Analysis')
        else:
            self.slope = slope

    def _set_base_load(self, base_load):
        """
        Sets base load attribute, checks for validity

        Parameters
        ----------
        base_load: float

        """
        if not np.isnan(base_load):
            self.base_load = base_load
        else:
            self.base_load = None

    def _calculate_regression_data(self):
        """
            Decide what data to use for the linear regression.
            In this case all data past the breakpoint
        """

        return self.df[self.df.independent > self.breakpoint]

    def _get_y(self, x):
        """
            Calculate a value on the trend line for a given x-value

            Parameters
            ----------
            x: number

            Returns
            -------
            float
        """
        # if there is an intersection and the value is before the intersection, return the baseload
        if self.intersect is not None and x <= self.intersect:
            return self.base_load
        else:
            # else return the normal value in the trendline
            return super(LinearRegression2, self)._get_y(x)

    def get_trend_x(self):
        """
            Return the values that are needed to draw the trendline.
            In this case, they are the same as the normal linear regression,
            but the intersection of the trend and the baseload need to be added
        """
        ret = super(LinearRegression2, self).get_trend_x()
        if self.intersect is not None:
            ret.append(self.intersect)
        return sorted(ret)

    def score_baseload(self, baseload_threshold=3, **kwargs):
        """
        Asign a score between 0 and 1 to the baseload, based on the number of points below the breakpoint.

        Parameters
        ----------
        baseload_threshold: int
            number of points below the breakpoint that make the baseline qualitative

        Returns
        -------
        float (between 0 and 1)
        """

        if self.intersect is not None and self.intersect > self.breakpoint:
            breakpoint = self.intersect
        else:
            breakpoint = self.breakpoint

        num_points = len(self.df[self.df.independent <= breakpoint])  # number of points below the breakpoint

        return min(1, float(num_points) / float(baseload_threshold))

    def score_total(self, weight=0.5, **kwargs):
        """
        Calculate score between 0 and 1 by combining all sub-scores

        Parameters
        ----------
        weight: float

        Returns
        -------
        float
        """
        score = super(LinearRegression2, self).score_total(weight=weight, **kwargs)
        baseload_score = self.score_baseload(**kwargs)

        return score * self._weigh_score(baseload_score, weight)


class LinearRegression3(LinearRegression2):
    """
    Calculate a linear regression that uses a predefined breakpoint to break between baseload and regression,
    yet exclude values in a certain range (percentage of the baseload) from the regression
    """

    def __init__(self, independent, dependent, breakpoint, percentage,
                 include_end_of_base_load=True, *args, **kwargs):
        """
        Parameters
        ----------
        independent : pandas.Series
        dependent : pandas.Series
        breakpoint : int | float
            point on the x-axis where to break between baseload and regression
        percentage: float
            y-values that are in this range to the baseload are excluded from the regression
        include_end_of_base_load : bool
        """

        self.percentage = percentage
        self.include_end_of_base_load = include_end_of_base_load
        super(LinearRegression3, self).__init__(independent=independent,
                                                dependent=dependent,
                                                breakpoint=breakpoint, *args,
                                                **kwargs)

    def _calculate_regression_data(self):
        """
            Decide what data to use for the linear regression.
            In this case all data past the breakpoint (from Linearregression2),
            but we iterate over them and drop values that are close to the base load.
        """
        # make a list of indices of entries that are to be excluded from the regression
        to_drop = self.df[self.df.independent <= self.breakpoint].sort_values(by='independent').index.tolist()

        # get the data from Regression 2
        res = super(LinearRegression3, self)._calculate_regression_data()

        if self.base_load is not None:
            # sort by x-value and iterate
            for entry in res.sort_values(by='independent').iterrows():
                # if the y value is smaller than the percentage of the baseload
                if entry[1].dependent < self.base_load * (1 + self.percentage):
                    # add the entry to the list to be dropped
                    to_drop.append(entry[0])
                else:
                    # if not, end the loop
                    break

            # if we want to include the last value of the base load in the regression, remove it from the todrop list
            if self.include_end_of_base_load and len(to_drop) > 0:
                to_drop.pop()

        # drop the to_drop list from the dataframe
        res = self.df.drop(to_drop)

        return res



import bs4
import requests
import iso8601
import datetime as dt
import pandas as pd
from .misc import dayset


def get_belpex(start, end=None):
    """
    Fetches Belpex prices for given period between start and end

    Parameters
    ----------
    start : something datetime like
    end : something datetime like too
        optional, defaults to tomorrow, because Belpex values are available for 1 day in the future!

    Returns
    -------
    Pandas Series
    """
    if end is None:
        end = dt.date.today() + dt.timedelta(days=1)
    # Prices are fetched per day, so we'll create a set of days, download seperate dataframes per day and put them
    # together
    series = [get_belpex_day(date) for date in dayset(start, end)]
    return pd.concat(series)


def get_belpex_day(date):
    """
    Get Belpex prices for a given day

    Parameters
    ----------
    date: something datetime like

    Returns
    -------
    Pandas series
    """
    # start out by getting the html from the website
    html = fetch_website(date)
    # parse html into an array of timestamps and values
    try:
        index, data = parse_html(html)
    except KeyError:  # something goes wrong when parsing, probably meaning data is unavailable.
        print("No data found for {}".format(date.date()))
        return None

    series = pd.Series(index=index, data=data)
    series = series.tz_convert('Europe/Brussels')
    return series


def fetch_website(date):
    """
    Fetch the website containing the data from ENTSOE.EU

    Parameters
    ----------
    date: something datetime-like

    Returns
    -------
    str
    """
    # the date needs to look like 24.02.2016 for the url
    formatted_date = date.strftime("%d.%m.%Y")
    url = "https://transparency.entsoe.eu/transmission-domain/r2/dayAheadPrices/show?name=&defaultValue=false&viewType=TABLE&areaType=BZN&atch=false&dateTime.dateTime={}+00:00|CET|DAY&biddingZone.values=CTY|10YBE----------2!BZN|10YBE----------2&dateTime.timezone=CET_CEST&dateTime.timezone_input=CET+(UTC+1)+/+CEST+(UTC+2)".format(
        formatted_date)

    r = requests.get(url)

    if r.status_code == 200:
        return r.text
    else:
        raise Exception("Seems like you got code {}".format(r.status_code))


def parse_html(html):
    """
    Searches the html-page for the elements we need

    Parameters
    ----------
    html : str

    Returns
    -------
    index : list of timestamps
    data : list of floats
    """
    index = []
    data = []

    # run the page through beautiful soup
    soup = bs4.BeautifulSoup(html, 'html.parser')

    # the elements we want are in a div with id dv-data-table
    data_table = soup.find(id='dv-data-table')
    # every element is in a span
    spans = data_table.findAll('span')
    for span in spans:
        # There is a iso8601 timestamp somewhere in an 'onclick' JavaScript call...
        # Luckily all arguments passed are of constant width so we can grab the stamp by index
        date = iso8601.parse_date(span['onclick'][128:152])
        index.append(date)
        # the value is right in the middle of the span
        value = float(span.text)
        data.append(value)

    return index, data



# -*- coding: utf-8 -*-
"""
Script to cache daily total elec, gas and water consumption for all sensors in the houseprint of OpenGrid

Created on 23/03/2017 by Roel De Coninck
"""

import os
import pandas as pd
from tqdm import tqdm

from opengrid.library import houseprint, caching
from opengrid import config

c = config.Config()

# Load houseprint from cache if possible, otherwise build it from source
try:
    hp_filename = os.path.join(c.get('data', 'folder'), 'hp_anonymous.pkl')
    hp = houseprint.load_houseprint_from_file(hp_filename)
    print("Houseprint loaded from {}".format(hp_filename))
except Exception as e:
    print(e)
    print("Because of this error we try to build the houseprint from source")
    hp = houseprint.Houseprint()

hp.init_tmpo()

# Get the cache objects for gas, elec and water, and update them, sensor by sensor
for sensortype in ['gas', 'electricity', 'water']:
    cache = caching.Cache(variable=sensortype + '_daily_total')
    sensors = hp.get_sensors(sensortype=sensortype)
    df_cached = cache.get(sensors=sensors)

    # for each sensor:
    # 1. get the last timestamp of the cached daily total
    # 2. get the daily data since than
    # 3. fill up the cache with the new data
    print('Caching daily totals for {}'.format(sensortype))
    for sensor in tqdm(sensors):
        try:
            last_ts = df_cached[sensor.key].dropna().index[-1]
            last_ts = last_ts.tz_convert('Europe/Brussels')
        except:
            last_ts = pd.Timestamp('1970-01-01', tz='Europe/Brussels')

        # Only get data until the end of the last day
        # Chunk the data retrieval and caching to avoid memory overflow
        end_ts = pd.Timestamp(pd.Timestamp('now', tz='Europe/Brussels'))
        months = pd.DatetimeIndex(start=last_ts, end=end_ts, freq='M').tolist()
        months.insert(0, last_ts - pd.Timedelta(days=2))
        months.append(end_ts)
        for start, end in zip(months[:-1], months[1:]):
            df = sensor.get_data(head=start - pd.Timedelta(days=1),
                                 tail=end,
                                 resample='day',
                                 diff=False,
                                 tz='Europe/Brussels')
            df = df.diff().shift(-1).dropna()
            if not len(df) == 0:
                cache.update(df)

    print("Updated {} with data from {} sensors".format(sensortype + '_daily_total', len(sensors)))





# coding: utf-8

# ## This script shows the visualization of electricity, water and gas consumption using carpet plots
# 
# To get started, first run the 'Synchronize data' script
# 
# #### Imports and paths

# In[ ]:

import os
import time
import pandas as pd


from opengrid import config
from opengrid.library import plotting
from opengrid.library import houseprint

c=config.Config()

try:
    if os.path.exists(c.get('tmpo', 'data')):
        path_to_tmpo_data = c.get('tmpo', 'data')
except:
    path_to_tmpo_data = None

# configuration for the plots
DEV = c.get('env', 'type') == 'dev' # DEV is True if we are in development environment, False if on the droplet
print("Environment configured for development: {}".format(DEV))
if not DEV:
    # production environment: don't try to display plots
    import matplotlib
    matplotlib.use('Agg')

import matplotlib.pyplot as plt
from matplotlib.dates import MinuteLocator, HourLocator, DateFormatter, AutoDateLocator, num2date

if DEV:
    if c.get('env', 'plots') == 'inline':
        get_ipython().magic(u'matplotlib inline')
    else:
        get_ipython().magic(u'matplotlib qt')
else:
    pass # don't try to render plots

plt.rcParams['figure.figsize'] = 16,8

# path to data
path_to_data = c.get('data', 'folder')
if not os.path.exists(path_to_data):
    raise IOError("Provide your path to the data in your config.ini file. ")
else:
    path_to_fig = os.path.join(path_to_data, 'figures')
    if not os.path.isdir(path_to_fig): os.makedirs(path_to_fig)


# In[ ]:

c.get('data','folder')


# ### Loading meta data and user variables

# In[ ]:

# Load houseprint from cache if possible, otherwise build it from source
try:
    hp_filename = os.path.join(c.get('data', 'folder'), 'hp_anonymous.pkl')
    hp = houseprint.load_houseprint_from_file(hp_filename)
    print("Houseprint loaded from {}".format(hp_filename))
except Exception as e:
    print(e)
    print("Because of this error we try to build the houseprint from source")
    hp = houseprint.Houseprint()

end = pd.Timestamp(time.time(), unit='s')
start = end - pd.Timedelta('21 days')


# ### Water sensors

# In[ ]:

water_sensors = hp.get_sensors(sensortype='water')
print("{} water sensors".format(len(water_sensors)))


# In[ ]:

for sensor in water_sensors:
    ts = sensor.get_data(head=start, tail=end)
    if not ts.dropna().empty:
        plotting.carpet(ts, title=' - '.join([sensor.device.key, sensor.description, sensor.key]), zlabel=r'Flow [l/min]')
        plt.savefig(os.path.join(path_to_fig, 'carpet_'+sensor.type+'_'+sensor.key), dpi=100)
        if not DEV:
            plt.close()


# ### Gas sensors

# In[ ]:

gas_sensors = hp.get_sensors(sensortype=('gas'))
print("{} gas sensors".format(len(gas_sensors)))


# In[ ]:

for sensor in gas_sensors:
    ts = sensor.get_data(head=start, tail=end)
    if not ts.dropna().empty:
        plotting.carpet(ts, title=' - '.join([sensor.device.key, sensor.description, sensor.key]), zlabel=r'Gas consumption [W]')
        plt.savefig(os.path.join(path_to_fig, 'carpet_'+sensor.type+'_'+sensor.key), dpi=100)
        if not DEV:
            plt.close()


# ### Electricity sensors

# In[ ]:

elec_sensors = hp.get_sensors(sensortype=('electricity'))
print("{} electricity sensors".format(len(elec_sensors)))


# In[ ]:

for sensor in elec_sensors:
    ts = sensor.get_data(head=start, tail=end)
    if not ts.dropna().empty:
        plotting.carpet(ts, title=' - '.join([sensor.device.key, sensor.description, sensor.key]), zlabel=r'Power [W]')
        plt.savefig(os.path.join(path_to_fig, 'carpet_'+sensor.type+'_'+sensor.key), dpi=100)
        if not DEV:
            plt.close()


# In[ ]:






import requests


class Slack(object):
    def __init__(self, url, username=None, channel=None, emoji=None):
        """
        Parameters
        ----------
        url : str
        username : str, optional
        channel : str, optional
        emoji : str, optional
        """
        self.url = url
        self.username = username
        self.channel = channel
        self.emoji = emoji

    def _post(self, p):
        """
        Parameters
        ----------
        p : dict
            payload

        Returns
        -------
        requests.Response
        """
        payload = p
        if self.username is not None:
            payload.update({"username": self.username})
        if self.channel is not None:
            payload.update({"channel": self.channel})
        if self.emoji is not None:
            payload.update({"icon_emoji": self.emoji})

        r = requests.post(url=self.url, json=payload, timeout=5)
        r.raise_for_status()
        return r

    def post_text(self, *text):
        """
        Parameters
        ----------
        text : str

        Returns
        -------
        requests.Response
        """
        payload = {"text": "\n".join(text)}
        return self._post(p=payload)

    def post_json(self, j):
        """
        Parameters
        ----------
        j : dict

        Returns
        -------
        requests.Response
        """
        return self._post(j)



__author__ = 'Jan Pecinovsky'

import pandas as pd

"""
A Site is a physical entity (a house, appartment, school, or other building).
It may contain multiple devices and sensors.
The Site contains most of the metadata, eg. the number of inhabitants, the size
of the building, the location etc.
"""


class Site(object):
    def __init__(self, hp=None, key=None, size=None, inhabitants=None,
                 postcode=None, construction_year=None, k_level=None,
                 e_level=None, epc_cert=None, tmpos=None):
        self.hp = hp  # backref to parent
        self.key = key
        self.size = size
        self.inhabitants = inhabitants
        self.postcode = postcode
        self.construction_year = construction_year
        self.k_level = k_level
        self.e_level = e_level
        self.epc_cert = epc_cert

        self.devices = []

        self._tmpos = tmpos

    @property
    def tmpos(self):
        if self._tmpos is not None:
            return self._tmpos
        elif self.hp.tmpos is not None:
            return self.hp.tmpos
        else:
            raise AttributeError('TMPO session not defined')

    @property
    def sensors(self):
        s = []
        for device in self.devices:
            for sensor in device.sensors:
                s.append(sensor)
        return s

    def __repr__(self):
        return """
    Site
    Key: {}
    {} devices
    {} sensors
    """.format(self.key,
               len(self.devices),
               len(self.sensors)
              )

    def get_sensors(self, sensortype = None):
        """
            Return a list with all sensors in this site

            Parameters
            ----------
            sensortype: gas, water, electricity: optional

            Returns
            -------
            list of Sensors
        """
        return [sensor for sensor in self.sensors if sensor.type == sensortype or sensortype is None]

    def get_data(self, sensortype=None, head=None, tail=None, diff='default', resample='min', unit='default'):
        """
        Return a Pandas Dataframe with the joined data for all sensors in this device

        Parameters
        ----------
        sensors : list of Sensor objects
            If None, use sensortype to make a selection
        sensortype : string (optional)
            gas, water, electricity. If None, and Sensors = None,
            all available sensors in the houseprint are fetched
        head, tail: timestamps,
        diff : bool or 'default'
            If True, the original data will be differentiated
            If 'default', the sensor will decide: if it has the attribute
            cumulative==True, the data will be differentiated.
        resample : str (default='min')
            Sampling rate, if any.  Use 'raw' if no resampling.
        unit : str , default='default'
            String representation of the target unit, eg m**3/h, kW, ...

        Returns
        -------
        Pandas DataFrame
        """
        sensors = self.get_sensors(sensortype)
        series = [sensor.get_data(head=head, tail=tail, diff=diff, resample=resample, unit=unit) for sensor in sensors]

        # workaround for https://github.com/pandas-dev/pandas/issues/12985
        series = [s for s in series if not s.empty]

        if series:
            df = pd.concat(series, axis=1)
        else:
            df = pd.DataFrame()

        # Add unit as string to each series in the df.  This is not persistent: the attribute unit will get
        # lost when doing operations with df, but at least it can be checked once.
        for s in series:
            try:
                df[s.name].unit = s.unit
            except:
                pass

        return df

    def add_device(self, device):
        """
        Parameters
        ----------
        device : Device
        """

        device.site = self
        self.devices.append(device)


# -*- coding: utf-8 -*-
"""
General analysis functions.

Try to write all methods such that they take a dataframe as input
and return a dataframe or list of dataframes.
"""

import datetime as dt
import pandas as pd


class Analysis(object):
    """
    Generic Analysis

    An analysis should have a dataframe as input
    self.result should be used as 'output dataframe'
    It also has output methods: plot, to json...
    """
    def __init__(self, df, *args, **kwargs):
        self.df = df
        self.do_analysis(*args, **kwargs)

    def do_analysis(self, *args, **kwargs):
        # To be overwritten by inheriting class
        self.result = self.df.copy()

    def plot(self):
        self.result.plot()

    def to_json(self):
        return self.result.to_json()


class DailyAgg(Analysis):
    """
    Obtain a dataframe with daily aggregated data according to an aggregation operator
    like min, max or mean
    - for the entire day if starttime and endtime are not specified
    - within a time-range specified by starttime and endtime.
      This can be used eg. to get the minimum consumption during the night.
    """
    def __init__(self, df, agg, starttime=dt.time.min, endtime=dt.time.max):
        """
        Parameters
        ----------
        df : pandas.DataFrame
            With pandas.DatetimeIndex and one or more columns
        agg : str
            'min', 'max', or another aggregation function
        starttime, endtime : datetime.time objects
            For each day, only consider the time between starttime and endtime
            If None, use begin of day/end of day respectively
        """
        super(DailyAgg, self).__init__(df, agg, starttime=starttime, endtime=endtime)

    def do_analysis(self, agg, starttime=dt.time.min, endtime=dt.time.max):
        if not self.df.empty:
            df = self.df[(self.df.index.time >= starttime) & (self.df.index.time < endtime)]
            df = df.resample('D', how=agg)
            self.result = df
        else:
            self.result = pd.DataFrame()















# -*- coding: utf-8 -*-
__author__ = 'Jan Pecinovsky'

import datetime as dt
import forecastio
from forecastio.models import Forecast
from geopy import Location, Point, GoogleV3
import numpy as np
import pandas as pd
import pytz
from cached_property import cached_property
from tqdm import tqdm
import os
import pickle

from .misc import dayset, calculate_temperature_equivalent, \
    calculate_degree_days
from opengrid import config
cfg = config.Config()


class Weather():
    """
        Object that contains Weather Data from Forecast.io for multiple days as a Pandas Dataframe.
        NOTE: Forecast.io allows 1000 requests per day, after that you have to pay. Each requested day is 1 request.
    """

    def __init__(self, location, start, end=None, cache=True):
        """
            Constructor

            Parameters
            ----------
            location : str | tuple(float, float)
                String can be City, address, POI. tuple = (lat, lng)
            start : datetime.datetime | pandas.Timestamp
                start of the interval to be searched
            end : datetime.datetime | pandas.Timestamp, optional
                end of the interval to be searched
                if None, use current time
            cache : bool
                use the cache or not
        """
        self.api_key = cfg.get('Forecast.io', 'apikey')
        self._location = location
        self._start = start
        self._end = end
        self.cache = cache

        self._forecasts = []

    @cached_property  # so we don't have to lookup every time it is called
    def location(self):
        """
        Get the location, but as a geopy location object

        Returns
        -------
        Location
        """
        # if the input was a string, we do a google lookup
        if isinstance(self._location, str):
            location = GoogleV3().geocode(self._location)

        # if the input was an iterable, it is latitude and longitude
        elif hasattr(self._location, '__iter__'):
            lat, long = self._location
            gepoint = Point(latitude=lat, longitude=long)
            location = Location(point=gepoint)

        else:
            raise ValueError('Invalid location')

        return location

    @property
    def start(self):
        """
        Return localized start time

        Returns
        -------
        datetime.datetime | pandas.Timestamp
        """
        if self._start.tzinfo is None:
            return self.tz.localize(self._start)
        else:
            return self._start

    @property
    def end(self):
        """
        Return localized end time

        Returns
        -------
        datetime.datetime | pandas.Timestamp
        """
        if self._end is None:
            return pd.Timestamp('now', tz=self.tz.zone)
        elif self._end.tzinfo is None:
            return self.tz.localize(self._end)
        else:
            return self._end

    @property
    def tz(self):
        """
        Get the local timezone of the requested location

        Returns
        -------
        pytz.timezone
        """
        # if there already are some forecasts, the timezone is in there
        if self._forecasts:
            tz = self._lookup_timezone()

        # use Google geocoder to lookup timezone
        else:
            lat, long, _alt = self.location.point
            tz = GoogleV3().timezone(location=(lat, long)).zone

        # return as a pytz object
        return pytz.timezone(tz)

    @property
    def forecasts(self):
        """
        Get a list of all forecast objects

        Returns
        -------
        list(Forecast)
        """
        # we stick the list in self._forecasts, only if it is still empty
        if not self._forecasts:
            # get list of seperate days
            days = dayset(start=self.start, end=self.end)
            # wrap in a tqdm so we get the progress bar
            for date in tqdm(days):
                self._forecasts.append(self._get_forecast(date=date))

        return self._forecasts

    def days(
            self,
            heating_base_temperatures=[16.5],
            cooling_base_temperatures=[18],
            irradiances=None,
            wind_orients=None
    ):
        """
        Create a dataframe with all weather data in daily resolution

        Parameters
        ----------
        heating_base_temperatures : list of numbers (optional, default 16.5)
            List of possible base temperatures for which to calculate heating degree days
        cooling_base_temperatures : list of numbers (optional, default 18)
            List of possible base temperatures for which to calculate cooling degree days
        irradiances : list[tuple], optional
            Calculate the global solar irradiance for multiple surfaces with a
            given orientation and tilt (in Wh/m^2)
            List with tuples. tuple = (orientation, tilt)
            Orientation in degrees from north (0 = north, 90 = east, ...)
            Tilt in degrees (0 = horizontal, 90 = vertical)
        wind_orients : list[int] | list[float], optional
            Calculate the wind force on surfaces with a given orientations
            Orientation in degrees from north (0 = north, 90 = east, ...)

        Returns
        -------
        pandas.DataFrame
        """

        # add 2 days before to calculate degree days
        self._add_forecast(self.start - pd.Timedelta(days=1))
        self._add_forecast(self.start - pd.Timedelta(days=2))

        # create a dataframe from the daily observations
        day_list = [self._forecast_to_day_series(forecast=forecast) for forecast in self.forecasts]
        frame = pd.concat(day_list)
        frame = self._fix_index(frame).sort_index()

        # add aggregates from hourly observations to the dataframe
        hourly_frame = self.hours(irradiances=irradiances,
                                  no_truncate=True,
                                  wind_orients=wind_orients)
        temperature = hourly_frame.temperature.resample('d').mean()
        ghi = hourly_frame.GlobalHorizontalIrradiance.dropna().resample('d').sum()
        tilted_gi = hourly_frame.filter(regex='^GlobalIrradiance').dropna().resample('d').sum()
        wind_force = hourly_frame.filter(regex='^windComponent').dropna().resample('d').sum()
        frame = pd.concat([frame, temperature, ghi, tilted_gi, wind_force], axis=1)
        frame = frame.tz_convert(self.tz.zone)  # because the concat loses tz info

        # add temperature equivalent and degree days
        frame['temperatureEquivalent'] = calculate_temperature_equivalent(temperatures=frame.temperature)
        frame.dropna(subset=['temperatureEquivalent'], inplace=True)

        for base in heating_base_temperatures:
            frame['heatingDegreeDays{}'.format(base)] = calculate_degree_days(
                temperature_equivalent=frame.temperatureEquivalent, base_temperature=base
            )

        for base in cooling_base_temperatures:
            frame['coolingDegreeDays{}'.format(base)] = calculate_degree_days(
                temperature_equivalent=frame.temperatureEquivalent, base_temperature=base, cooling=True
            )

        return frame

    def hours(self, irradiances=None, no_truncate=False, wind_orients=None):
        """
        Create a dataframe with all weather data in hourly resolution

        Parameters
        ----------
        irradiances : list[tuple], optional
            Calculate the global solar irradiance for multiple surfaces with a
            given orientation and tilt
            List with tuples. tuple = (orientation, tilt)
            Orientation in degrees from north (0 = north, 90 = east, ...)
            Tilt in degrees (0 = horizontal, 90 = vertical)
        no_truncate : bool
            do not truncate to the exact timestamps
            useful for when using the hour dataframe to do aggregations
        wind_orients : list[int] | list[float], optional
            Calculate the wind force on surfaces with a given orientations
            Orientation in degrees from north (0 = north, 90 = east, ...)

        Returns
        -------
        pandas.DataFrame
        """
        day_list = [self._forecast_to_hour_series(forecast) for forecast in self.forecasts]
        frame = pd.concat(day_list)
        frame = self._fix_index(frame)
        frame.sort_index(inplace=True)
        if not no_truncate:
            frame = frame.truncate(before=self.start, after=self.end)

        if irradiances is not None:
            for ir in irradiances:
                frame = self._add_irradiance(frame=frame, orient=ir[0], tilt=ir[1])

        if wind_orients is not None:
            for orient in wind_orients:
                frame = self._add_wind_components(frame=frame, orient=orient)

        return frame

    def _get_forecast(self, date):
        """
            Get the raw forecast object for a given date

            Parameters
            ----------
            date : datetime.date

            Returns
            -------
            forecastio forecast
        """
        # Forecast takes a dt.datetime
        # conversion from dt.date to dt.datetime, there must be a better way, right?
        time = dt.datetime(year=date.year, month=date.month, day=date.day)

        f = None
        if self.cache:
            f = self._load_from_cache(date)

        if not f:
            # We specifically ask for si units,
            # so we can inject the SOLAR argument to get solar data
            # which is in beta (januari 2017)
            f = forecastio.load_forecast(key=self.api_key,
                                         lat=self.location.latitude,
                                         lng=self.location.longitude,
                                         time=time,
                                         units='si&solar'
                                         )
            if self.cache:
                self._save_in_cache(f, date)

        return f

    def _get_forecast_dates(self):
        """
        Return a set with the dates of all forecasts in self.forecasts

        Returns
        -------
        set of datetime.date
        """
        dates = []
        for forecast in self.forecasts:
            time = forecast.currently().time

            # the time is in UTC, we need to localize it.
            tz = pytz.timezone(self._lookup_timezone())
            time_utc = tz.fromutc(time)

            dates.append(time_utc.date())
        return set(dates)

    def _add_forecast(self, date):
        """
        Add a forecast to the list of forecasts

        Parameters
        ----------
        date : dt.date | dt.datetime | pd.Timestamp
        """
        # for if you pass a datetime instead of a date
        if hasattr(date, 'date'):
            date = date.date()

        if date not in self._get_forecast_dates():
            self._forecasts.append(self._get_forecast(date))

    def _forecast_to_hour_series(self, forecast):
        """
        Transforms the hourly data of a forecast object to a pandas dataframe

        Parameters
        ----------
        forecast : Forecast object

        Returns
        -------
        Pandas Dataframe

        """

        hour_list = [pd.Series(self._flatten_solar(hour.d)) for hour in forecast.hourly().data]
        frame = pd.concat(hour_list, axis=1).T
        frame.temperature = frame.temperature.astype(float)
        return frame

    @staticmethod
    def _flatten_solar(j):
        """
        Extracts the 'solar' part from the response json,
        renames the fields and includes them along with everything else

        Parameters
        ----------
        j : dict

        Returns
        -------
        dict
        """
        try:
            solar = j.pop('solar')
        except KeyError:
            return j
        else:
            # give the properties some more verbose names
            new_names = {'altitude': 'SolarAltitude',
                         'dni': 'DirectNormalIrradiance',
                         'ghi': 'GlobalHorizontalIrradiance',
                         'dhi': 'DiffuseHorizontalIrradiance',
                         'etr': 'ExtraTerrestrialRadiation',
                         'azimuth': 'SolarAzimuth'
                         }
            solar = {new_names[key]: val for key, val in solar.items()}
            j.update(solar)

            # workaround for the 90 bug by Dark Sky
            # add 90 and take modulo 360 to stay between 0 and 360
            j.update({'SolarAzimuth': (j.get('SolarAzimuth') + 90) % 360})

            return j

    def _fix_index(self, frame):
        """
        Transform a column of a dataframe to a datetimeindex, set as the index and localize

        Parameters
        ----------
        frame : Pandas Dataframe

        Returns
        -------
        Pandas Dataframe

        """
        frame['time'] = pd.DatetimeIndex(frame['time'].astype('datetime64[s]'))
        frame.set_index('time', inplace=True)
        frame = frame.tz_localize('UTC')
        frame = frame.tz_convert(self.tz.zone)
        return frame

    def _lookup_timezone(self):
        """
        Lookup the timezone in the JSON of the first forecast

        Returns
        -------
        String (Pytz timezone)
        """
        tz = self.forecasts[0].json['timezone']
        return tz

    def _forecast_to_day_series(self, forecast):
        """
        Transforms the daily data of a forecast object to a pandas dataframe

        Parameters
        ----------
        forecast : forecastio.models.Forecast

        Returns
        -------
        pandas.DataFrame

        """
        data = forecast.daily().data[0].d
        frame = [pd.Series(data)]
        return pd.concat(frame, axis=1).T

    @property
    def cache_folder(self):
        location_str = "{}_{}".format(round(self.location.latitude, 4),
                                      round(self.location.longitude, 4))

        forecast_folder = os.path.join(os.path.abspath(cfg.get('data', 'folder')),
                            'forecasts')
        location_folder = os.path.join(forecast_folder, location_str)

        for folder in [forecast_folder, location_folder]:
            if not os.path.exists(folder):
                print("This folder does not exist: {}, it will be created".format(folder))
                os.mkdir(folder)

        return location_folder

    def _pickle_path(self, date):
        filename = str(date) + '.pkl'
        path = os.path.join(self.cache_folder, filename)
        return path

    def _save_in_cache(self, f, date):
        """
        Save Forecast object to cache

        Parameters
        ----------
        f : Forecast
        date : datetime.date
        """

        pickle.dump(f, open(self._pickle_path(date), "wb"))

    def _load_from_cache(self, date):
        """
        Load Forecast object from cache

        Parameters
        ----------
        date : datetime.date

        Returns
        -------
        Forecast
        """
        path = self._pickle_path(date)
        if os.path.exists(path):
            return pickle.load(open(path, "rb"))
        else:
            return None

    @staticmethod
    def irradiance_on_tilted_surface(dni, dhi, altitude, azimuth, orient, tilt):
        """
        Calculate the global solar radiation on a tilted surface

        Parameters
        ----------
        dni : pandas.Series
            Direct Normal Irradiance
        dhi : pandas.Series
            Diffuse Horizontal Irradiance
        altitude : pandas.Series
            altitude of the sun above the horizon (in degrees)
            0 = horizon, 90 = right above
        azimuth : pandas.Series
            location of the sun in respect to North (in degrees)
            0 = North, 90 = East, 180 = South, 270 = West
        orient : int|float
            orientation of the surface
            0 = North, 90 = East, 180 = South, 270 = West
        tilt : int|float
            tilt of the surface
            0 = Horizontal, 90 = Vertical

        Returns
        -------
        pandas.Series
        """
        a = np.radians(altitude.astype(float))
        b = np.radians(float(tilt))
        c = np.radians(float(orient))
        d = np.radians(azimuth.astype(float))

        direct = dni * (np.cos(a) * np.sin(b) * np.cos(c - d) + np.sin(a) * np.cos(b))
        direct[direct < 0] = 0
        return direct + dhi

    def _add_irradiance(self, frame, orient, tilt):
        """
        Add a column to the frame with the global irradiance on a surface for
        a given orientation and tilt

        Parameters
        ----------
        frame : pandas.DataFrame
        orient : float|int
            in degrees from north
        tilt : float|int
            in degrees from horizontal

        Returns
        -------
        pandas.DataFrame
        """
        new_col = self.irradiance_on_tilted_surface(
            dni=frame.DirectNormalIrradiance,
            dhi=frame.DiffuseHorizontalIrradiance,
            altitude=frame.SolarAltitude,
            azimuth=frame.SolarAzimuth,
            orient=orient,
            tilt=tilt
        )
        name = "GlobalIrradianceO{}T{}".format(int(orient), int(tilt))

        frame[name] = new_col

        return frame

    @staticmethod
    def wind_on_oriented_face(bearing, speed, orient):
        """
        Calculate the wind speed with respect to a given orientation
        (of a building wall for example)

        Parameters
        ----------
        bearing : pandas.Series
        speed : pandas.Series
        orient : int | float

        Returns
        -------
        pandas.Series
        """
        b = np.radians(bearing.astype(float))
        o = np.radians(float(orient))

        wind = speed * np.cos(b - o)
        wind[wind < 0] = 0
        return wind

    def _add_wind_components(self, frame, orient):
        """
        Add columns to the frame with
            the wind speed,
            the wind speed squared (force or pressure)
            the wind speed cubed (power on a turbine)
        for a given orientation

        Parameters
        ----------
        frame : pandas.DataFrame
        orient : float|int
            in degrees from north

        Returns
        -------
        pandas.DataFrame
        """
        oriented_speed = self.wind_on_oriented_face(
            bearing=frame.windBearing,
            speed=frame.windSpeed,
            orient=orient
        )

        name = "windComponent{}".format(int(orient))
        frame[name] = oriented_speed

        name = "windComponentSquared{}".format(int(orient))
        frame[name] = oriented_speed ** 2

        name = "windComponentCubed{}".format(int(orient))
        frame[name] = oriented_speed ** 3

        return frame




# coding: utf-8

# In[ ]:

# opengrid imports
from opengrid.library import misc, houseprint, caching
from opengrid.library.analysis import DailyAgg
from opengrid import config
from opengrid.library.slack import Slack
from opengrid.library import alerts
c=config.Config()

# other imports
import pandas as pd
import json
import charts
import numpy as np
import os
import datetime as dt
import pytz
BXL = pytz.timezone('Europe/Brussels')


# In[ ]:

# Load houseprint from cache if possible, otherwise build it from source
try:
    hp_filename = os.path.join(c.get('data', 'folder'), 'hp_anonymous.pkl')
    hp = houseprint.load_houseprint_from_file(hp_filename)
    print("Houseprint loaded from {}".format(hp_filename))
except Exception as e:
    print(e)
    print("Because of this error we try to build the houseprint from source")
    hp = houseprint.Houseprint()

sensors = hp.get_sensors(sensortype='electricity') # sensor objects

# Remove some sensors
exclude = [
            '565de0a7dc64d8370aa321491217b85f' # 3E
          ]
solar = [x.key for x in hp.search_sensors(type='electricity', system='solar')]
exclude += solar

for s in sensors:
    if s.key in exclude:
        sensors.remove(s)

hp.init_tmpo()

# In[ ]:


sensors = hp.get_sensors() # sensor objects
t = []
for s in sensors:
    key = s.key
    lt = s.last_timestamp()
    if lt is None:
        continue
    nr_of_days = pd.Timestamp('now', tz='UTC') - lt
    t.append((key,nr_of_days))


# In[ ]:

df = pd.DataFrame(t).rename(columns={0: "sensor_id", 1: "timedelta"}).set_index('sensor_id')


# In[ ]:

df['seconds'] = df.timedelta.map(lambda x: x.total_seconds())
df['days'] = df.timedelta.map(lambda x: x.days)


# # Setup NoDataBot slack bot

# In[ ]:

slack_url = c.get('Slack', 'webhook')
username = 'NoDataBot'
channel = "junk" # we don't want to clutter up everything
emoji = ':warning:'
title = 'No data for sensor'
description = 'We have not found data for the last 24 hours (= 86400 seconds) for the following sensor. The result is expressed in seconds.'

slack = Slack(url=slack_url, username=username, channel=channel, emoji=emoji)


# # Create the alerts and send

# In[ ]:

alerts.create_alerts(df, hp, 'no_sensor_data', slack, title, description, column='seconds')



"""
Create alerts

"""

import json
from opengrid import config
c=config.Config()
import os

def get_threshold(analysis, sensor_key):
    """
    Return threshold value for a given sensor_key for a given analysis.
    The threshold is fetched from the alerts.cfg file.

    Parameters
    ----------
    analysis : str
        Name of the analysis
    sensor_key : str
        Sensor key for which the threshold has to be returned.
    """
    path_alerts = c.get('Slack', 'config')
    threshold = json.load(open(path_alerts, "r"))
    default = threshold[analysis]['default']

    return threshold[analysis].get(sensor_key, default)


def create_alerts(df, hp, analysis, slack, title,  description, column='result'):
    """
    Create alerts for each sensor, if needed.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with specific structure:
            - sensor_keys as index
            - by default, needs a column called 'result' with the data to be compared to a threshold
            You can specify a different column
    analysis : str
        Name of the analysis
    slack : Slack object
        See library.slack.py
    column : str, default='result'
        Column name with the results to be used for the alerting
    """

    for sensor_key in df.index:
        tr = get_threshold(analysis, sensor_key=sensor_key)
        if df.loc[sensor_key, column] > tr:
            json_message = {
                "text": "",
                "attachments": [
                    {
                        "title": title,
                        "text": description,
                        "fallback": "OpenGrid alert",
                        "color": "warning",  # this will create a red line
                        "fields": [
                            {
                                "title": "Fluksometer",
                                "value": hp.find_sensor(sensor_key).device.key,
                                "short": True
                            },
                            {
                                "title": "Sensor key",
                                "value": sensor_key,
                                "short": True
                            },
                            {
                                "title": "Your result",
                                "value": df.loc[sensor_key, column],
                                "short": True
                            },
                            {
                                "title": "Your threshold value",
                                "value": tr,
                                "short": True
                            },
                           {
                                "title": "Link to OpenGrid website",
                                "value": "https://opengrid.be/sensor/" + sensor_key,
                                "short": True
                            }
                        ]
                    }
                ]
            }
            slack.post_json(json_message)



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 30 02:37:25 2013

@author: roel
"""

import os
import sys
import unittest
import inspect
import numpy as np
import pytz

test_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# add the path to OpenGrid to sys.path
sys.path.insert(1, os.path.join(test_dir, os.pardir, os.pardir, os.pardir))
from opengrid.library.misc import *


class MiscTest(unittest.TestCase):
    """
    Class for testing the misc functions
    """

    def test_parse_date_from_datetime(self):
        """Parsing a datetime into a pandas.Timestamp"""

        bxl = pytz.timezone('Europe/Brussels')
        dt_ = bxl.localize(dt.datetime(2014, 11, 23, 1, 2, 3))
        epoch = pytz.UTC.localize(dt.datetime(1970, 1, 1, 0, 0, 0))
        epoch_expected = (dt_ - epoch).total_seconds()

        pts = parse_date(dt_)
        self.assertEqual(pts.value / 1e9, epoch_expected)

    def test_parse_date_from_datetime_naive(self):
        """Parsing a naïve datetime into a pandas.Timestamp makes it UTC"""

        dt_ = pytz.UTC.localize(dt.datetime(2014, 11, 23, 1, 2, 3))
        epoch = pytz.UTC.localize(dt.datetime(1970, 1, 1, 0, 0, 0))
        epoch_expected = (dt_ - epoch).total_seconds()

        pts = parse_date(dt.datetime(2014, 11, 23, 1, 2, 3))
        self.assertEqual(pts.value / 1e9, epoch_expected)

    def test_parse_date_from_posix(self):
        """Parsing a float"""

        pts = parse_date(1416778251.460574)
        self.assertEqual(1416778251.460574, pts.value / 1e9)

    def test_parse_date_from_string(self):
        """Parsing some commong types of strings"""

        dt_ = pytz.UTC.localize(dt.datetime(2014, 11, 23, 1, 2, 3))
        epoch = pytz.UTC.localize(dt.datetime(1970, 1, 1, 0, 0, 0))
        epoch_expected = (dt_ - epoch).total_seconds()

        pts = parse_date('20141123 01:02:03')
        self.assertEqual(pts.value / 1e9, epoch_expected)

        pts = parse_date('2014-11-23 01:02:03')
        self.assertEqual(pts.value / 1e9, epoch_expected)

        pts = parse_date('2014-11-23T010203')
        self.assertEqual(pts.value / 1e9, epoch_expected)

    def test_time_to_timedelta(self):
        t = dt.time(2, 45, 23)
        self.assertEqual(time_to_timedelta(t).total_seconds(), 7200 + 45 * 60 + 23.0)

        t = dt.time(2, 45, 23, tzinfo=pytz.timezone('Europe/Brussels'))
        self.assertEqual(time_to_timedelta(t).total_seconds(), 7200 + 45 * 60 + 23.0)

    def test_split_by_day(self):
        index = pd.DatetimeIndex(start='20160101 03:51:15', freq='h', periods=80)
        df = pd.DataFrame(index=index, data=np.random.randn(80, 2), columns=['A', 'B'])

        # without specifying hours
        list_daily = split_by_day(df)
        self.assertEqual(len(list_daily), 4)
        self.assertEqual(list_daily[0].index[0], index[0])
        self.assertEqual(list_daily[1].index[0], pd.Timestamp('20160102 00:51:15'))
        self.assertEqual(list_daily[1].index[-1], pd.Timestamp('20160102 23:51:15'))

        # specifying start
        list_daily = split_by_day(df, starttime=dt.time(1, 30))
        self.assertEqual(len(list_daily), 4)
        self.assertEqual(list_daily[0].index[0], pd.Timestamp('20160101 03:51:15'))
        self.assertEqual(list_daily[1].index[0], pd.Timestamp('20160102 01:51:15'))
        self.assertEqual(list_daily[1].index[-1], pd.Timestamp('20160102 23:51:15'))

        # specifying start and end
        list_daily = split_by_day(df, starttime=dt.time(1, 30), endtime=dt.time(6))
        self.assertEqual(len(list_daily), 4)
        self.assertEqual(list_daily[0].index[0], pd.Timestamp('20160101 03:51:15'))
        self.assertEqual(list_daily[1].index[0], pd.Timestamp('20160102 01:51:15'))
        self.assertEqual(list_daily[1].index[-1], pd.Timestamp('20160102 05:51:15'))

    def test_unit_conversion_factor(self):
        cf = unit_conversion_factor('liter/minute', 'm**3/hour')
        np.testing.assert_array_almost_equal(cf, 1 / 1e3 * 60.)

        cf = unit_conversion_factor('Wh/min', 'kW')
        np.testing.assert_array_almost_equal(cf, 60 / 1000.)

    def test_dayset(self):
        ds = dayset(start=pd.Timestamp('20160101'), end=pd.Timestamp('20160131'))
        comp = [dt.date(year=2016, month=1, day=day) for day in range(1, 32)]
        self.assertEqual(ds, comp)

    def test_split_irregular_date_list(self):
        d1 = dayset(start=pd.Timestamp('20160401'), end=pd.Timestamp('20160405'))
        d2 = dayset(start=pd.Timestamp('20160101'), end=pd.Timestamp('20160103'))
        date_list = d1 + d2
        split = split_irregular_date_list(date_list=date_list)
        self.assertEqual(split[0][0], dt.date(year=2016, month=1, day=1))
        self.assertEqual(split[0][1], dt.date(year=2016, month=1, day=3))
        self.assertEqual(split[1][0], dt.date(year=2016, month=4, day=1))
        self.assertEqual(split[1][1], dt.date(year=2016, month=4, day=5))

    def test_calculate_temperature_equivalent(self):
        temps = [8.3, 8.7, 9.2]
        t_equiv = calculate_temperature_equivalent(pd.Series(temps))
        last = t_equiv.iloc[-1]
        last_man = 0.6*temps[2] + 0.3*temps[1] + 0.1*temps[0]
        self.assertEqual(last, last_man)
        self.assertEqual(t_equiv.name, 'temp_equivalent')

    def test_calculate_degree_days(self):
        temp_equivs = [-5.0, 1.5, 25.5]
        hdd = calculate_degree_days(temperature_equivalent=pd.Series(temp_equivs), base_temperature=16.5)
        self.assertEqual(hdd.tolist(), [21.5, 15.0, 0.0])
        self.assertEqual(hdd.name, 'heating_degree_days_16.5')

        cdd = calculate_degree_days(temperature_equivalent=pd.Series(temp_equivs), base_temperature=24, cooling=True)
        self.assertEqual(cdd.tolist(), [0.0, 0.0, 1.5])
        self.assertEqual(cdd.name, 'cooling_degree_days_24')


if __name__ == '__main__':
    # http://stackoverflow.com/questions/4005695/changing-order-of-unit-tests-in-python
    if sys.version_info.major >= 3:  # compatibility python 3
        ln = lambda f: getattr(MiscTest, f).__code__.co_firstlineno  # functions have renamed attributes
        lncmp = lambda _, a, b: (ln(a) > ln(b)) - (
        ln(a) < ln(b))  # cmp() was deprecated, see https://docs.python.org/3.0/whatsnew/3.0.html
    else:
        ln = lambda f: getattr(MiscTest, f).im_func.func_code.co_firstlineno
        lncmp = lambda _, a, b: cmp(ln(a), ln(b))

    unittest.TestLoader.sortTestMethodsUsing = lncmp

    suite1 = unittest.TestLoader().loadTestsFromTestCase(MiscTest)
    alltests = unittest.TestSuite([suite1])

    # selection = unittest.TestSuite()
    # selection.addTest(HouseprintTest('test_get_sensor'))

    unittest.TextTestRunner(verbosity=1, failfast=False).run(alltests)



import unittest
from six import string_types
from pandas.core.series import Series

from opengrid.library.belpex import *


class BelpexTest(unittest.TestCase):
    """
    Class for testing the belpex web scraper
    """

    def setUp(self):
        self.website = fetch_website(date=dt.datetime.today())

    def test_fetch_website(self):
        self.assertIsInstance(self.website, string_types)

    def test_parse_html(self):
        parsed = parse_html(self.website)
        self.assertIsInstance(parsed, tuple)

        dates, values = parsed
        self.assertIsInstance(dates, list)
        self.assertIsInstance(values, list)
        self.assertEqual(len(dates), len(values))
        self.assertIsInstance(dates[0], dt.datetime)
        self.assertIsInstance(values[0], float)

        website_err = fetch_website(date=dt.datetime.today() + dt.timedelta(days=5))
        with self.assertRaises(KeyError):
            parse_html(website_err)

    def test_get_belpex_day(self):
        series = get_belpex_day(date=dt.datetime.today())
        self.assertIsInstance(series, Series)

        self.assertIsNone(get_belpex_day(date=dt.datetime.today() + dt.timedelta(days=5)))

    def test_get_belpex(self):
        series = get_belpex(start=dt.datetime.today() - dt.timedelta(days=1), end=dt.datetime.today())
        self.assertIsInstance(series, Series)


if __name__ == '__main__':
    unittest.main()



__author__ = 'Jan Pecinovsky'

import geocoder
import astral
import math
import pandas as pd

class SolarInsolation(object):
    """
        Module to calculate Solar Insolation (direct intensity, global intensity, air mass) and
        basic solar parameters (angle) based on a location and a date.
        Formulas from pveducation.org
    """
    def __init__(self, location):
        """
            Parameters
            ----------
            location: String
        """
        self.location = geocoder.google(location)
        self.elevation = geocoder.google(self.location.latlng, method='Elevation').elevation
        self.astral = astral.Astral()

    def _airMass(self, angleFromVertical):
        """
            Raw formula to calculate air mass

            Parameters
            ----------
            angleFromVertical: float
                in radians

            Returns
            -------
            float
        """
        denom = math.cos(angleFromVertical) + 0.50572 * (96.07995 - angleFromVertical) ** -1.6364
        if denom >= 0:
            return 1 / denom
        else:
            return -1

    def airMass(self, datetime):
        """
            Calculate air mass for a given date and time

            Parameters
            ----------
            datetime: datetime.datetime

            Returns
            -------
            float
        """
        angleFromVertical = (math.pi/2) - self.solarElevation(datetime)
        return self._airMass(angleFromVertical)

    def _directIntensity(self, elevation, airMass):
        """
            Raw Formula to calculate direct solar beam intensity

            Parameters
            ---------
            elevation: float
                in meters
            airMass: float

            Returns
            -------
            float
                in W/m**2
        """
        elevation = elevation / 1000 #formula uses km
        di = 1.353 * ((1 - 0.14*elevation) * 0.7**airMass**0.678 + 0.14*elevation)
        return di*1000 #formula output is kW/m**2

    def directIntensity(self, datetime):
        """
            Calculate direct solar beam intensity for a given date and time

            Parameters
            ----------
            datetime: datetime.datetime

            Returns
            -------
            float
                in W/m**2
        """
        airMass = self.airMass(datetime)
        if airMass == -1:
            return 0
        return self._directIntensity(self.elevation, airMass)

    def _backgroundIrradiance(self, directIntensity):
        """
        Calculate the background irradiance, which is 10% of the direct intensity
        :param directIntensity: float
        :return: float
        """

        return 0.1 * directIntensity

    def _globalIrradiance(self, directIntensity):
        """
            Raw formula to calculate global solar irradiance

            Parameters
            ----------
            directIntensity: float
                in W/m**2

            Returns
            -------
            float
                in W/m**2
        """
        return directIntensity + self._backgroundIrradiance(directIntensity)

    def globalIrradiance(self, datetime):
        """
            Calculate global solar irradiance for a given date and time

            Parameters
            ----------
            datetime: datetime.datetime

            Returns
            -------
            float
                in W/m**2
        """
        directIntensity = self.directIntensity(datetime)
        return self._globalIrradiance(directIntensity)

    def solarElevation(self, datetime):
        """
            Calculate angle of the sun above the horizon

            Parameters
            ----------
            datetime: datetime.datetime

            Returns
            -------
            float
                in radians
        """
        deg = self.astral.solar_elevation(dateandtime = datetime,
                                          latitude = self.location.lat,
                                          longitude = self.location.lng)
        return math.radians(deg)

    def df(self, start, end):
        """
            Creates a dataframe with the insolation in W/m**2 in hourly resolution

            Parameters
            ----------
            start, end: datetime.datetime

            Returns
            -------
            Pandas Dataframe
        """

        hours = pd.date_range(start=start, end=end, freq='h')
        gis = []
        for hour in hours:
            gis.append(self.globalIrradiance(hour))

        df = pd.DataFrame(gis, index=hours, columns = ['insolation'])
        return df.tz_localize('UTC')

    def solarAzimuth(self, datetime):
        """
        Calculate the azimuth of the sun

        :param datetime: datetime.datetime
        :return: float
            in radians
        """

        deg = self.astral.solar_azimuth(dateandtime=datetime,
                                        latitude=self.location.lat,
                                        longitude=self.location.lng)
        return math.radians(deg)

class PVModel(SolarInsolation):
    """
        Module that models a theoretically perfect PV installation,
        extending the Solar Insolation Model, but adding PV orientation and tilt.
    """

    def __init__(self, location, orient=180, tilt=35):
        """
            Parameters
            ----------
            location: String
            orient: number (optional, default=180 (south))
                degrees (0-360)
            tilt: number (optional, default=35)
                degrees
        """

        super(PVModel, self).__init__(location = location)
        self.orient = math.radians(orient)
        self.tilt = math.radians(tilt)

    def directIntensity(self, datetime):
        """
            Calculate the direct solar beam intensity for a given date and time,
            but compensate for the PV orientation and tilt

            Parameters
            ----------
            datetime: datetime.datetime

            Returns
            -------
            float
                in W/m**2
        """
        di = super(PVModel, self).directIntensity(datetime)
        a = self.solarElevation(datetime)
        b = self.tilt
        c = self.orient
        d = self.solarAzimuth(datetime)

        PVdi = di * (math.cos(a)*math.sin(b)*math.cos(c-d) + math.sin(a)*math.cos(b))

        #PVdi cannot be negative
        return max(0, PVdi)

    def globalIrradiance(self, datetime):
        """
            Calculate global solar irradiance for a given date and time
            needed to override, because background irradiance is not influenced by PV orientation and tilt

            Parameters
            ----------
            datetime: datetime.datetime

            Returns
            -------
            float
                in W/m**2
        """
        #calculate the direct intensity without influence of tilt and orientation
        di = super(PVModel, self).directIntensity(datetime)

        #add the tilted and oriented direct intensity to the background irradiance
        return self.directIntensity(datetime) + self._backgroundIrradiance(di)


# -*- coding: utf-8 -*-
"""
opengrid

The config module provides support for handling configuration
such as the location to store files, etc.

The configuration is composed from different sources in a
hierarchical fashion:
    * The defaults hard-coded in this file.
    * If it exists: the opengrid.cfg file in the directory
      of this module
    * If it exists: the opengrid.cfg file in the 'current'
      directory
    * If provided: the file given to the constructor
The lower a source is in the list, to more it has precedence.

Not all files must contain all configuration properties. A
configuration file can also overwrite only a subset of the
configuration properties.
"""
import os, sys

#configparser changed name in Python3
if sys.version_info[0] == 3:
    from configparser import SafeConfigParser
else:
    from ConfigParser import SafeConfigParser

import inspect


class Config(SafeConfigParser):
    """
    The config class implements the ConfigParser interface.
    More specifically, it inherits from SafeConfigParser.

    See the documentation of SafeConfigParser for the available
    methods.
    """

    def __init__(self, configfile=None):
        SafeConfigParser.__init__(self)
        self.__add_defaults()
        configfiles = []
        # Add the filename for the config file in the modules
        # directory
        self.opengrid_libdir = os.path.dirname(os.path.abspath(
            inspect.getfile(inspect.currentframe())))
        configfiles.append(os.path.join(self.opengrid_libdir, 'opengrid.cfg'))
        # Add the filename for the config file in the 'current' directory
        configfiles.append('opengrid.cfg')
        # Add the filename for the config file passed in
        if configfile:
            configfiles.append(configfile)
        self.read(configfiles)

    def __add_defaults(self):
        self.add_section('opengrid_server')
        self.set('opengrid_server', 'host', '95.85.34.168')
        self.set('opengrid_server', 'port', '8080')
        self.set('opengrid_server', 'user', 'opengrid')
        self.set('opengrid_server', 'password', 'CHANGE ME IN AN OPENGRID.CFG FILE')

        self.add_section('data')
        self.set('data', 'folder', os.path.expanduser("~/.opengrid/data"))

        self.add_section('houseprint')
        self.set('houseprint', 'password', 'CHANGE ME IN AN OPENGRID.CFG FILE')

        self.add_section('env')
        self.set('env', 'type', 'dev')
        self.set('env', 'plots', 'inline')









# -*- coding: utf-8 -*-
"""
Created on Wed Nov 26 18:03:24 2014

@author: KDB
"""
import numpy as np
import pandas as pd
import datetime as dt
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.dates import date2num, num2date, HourLocator, DayLocator, AutoDateLocator, DateFormatter
from matplotlib.colors import LogNorm


def carpet(timeseries, **kwargs):
    """
    Draw a carpet plot of a pandas timeseries.

    The carpet plot reads like a letter. Every day one line is added to the
    bottom of the figure, minute for minute moving from left (morning) to right
    (evening).
    The color denotes the level of consumption and is scaled logarithmically.
    If vmin and vmax are not provided as inputs, the minimum and maximum of the
    colorbar represent the minimum and maximum of the (resampled) timeseries.

    Parameters
    ----------
    timeseries : pandas.Series
    vmin, vmax : If not None, either or both of these values determine the range
    of the z axis. If None, the range is given by the minimum and/or maximum
    of the (resampled) timeseries.
    zlabel, title : If not None, these determine the labels of z axis and/or
    title. If None, the name of the timeseries is used if defined.
    cmap : matplotlib.cm instance, default coolwarm
    """

    # define optional input parameters
    cmap = kwargs.pop('cmap', cm.coolwarm)
    norm = kwargs.pop('norm', LogNorm())
    interpolation = kwargs.pop('interpolation', 'nearest')
    cblabel = kwargs.pop('zlabel', timeseries.name if timeseries.name else '')
    title = kwargs.pop('title', 'carpet plot: ' + timeseries.name if timeseries.name else '')

    # data preparation
    if timeseries.dropna().empty:
        print('skipped {} - no data'.format(title))
        return
    ts = timeseries.resample('min', how='mean', label='left', closed='left')
    vmin = max(0.1, kwargs.pop('vmin', ts[ts > 0].min()))
    vmax = max(vmin, kwargs.pop('vmax', ts.quantile(.999)))

    # convert to dataframe with date as index and time as columns by
    # first replacing the index by a MultiIndex
    # tz_convert('UTC'): workaround for https://github.com/matplotlib/matplotlib/issues/3896
    mpldatetimes = date2num(ts.index.tz_convert('UTC').astype(dt.datetime))
    ts.index = pd.MultiIndex.from_arrays(
        [np.floor(mpldatetimes), 2 + mpldatetimes % 1])  # '2 +': matplotlib bug workaround.
    # and then unstacking the second index level to columns
    df = ts.unstack()

    # data plotting

    fig, ax = plt.subplots()
    # define the extent of the axes (remark the +- 0.5  for the y axis in order to obtain aligned date ticks)
    extent = [df.columns[0], df.columns[-1], df.index[-1] + 0.5, df.index[0] - 0.5]
    im = plt.imshow(df, vmin=vmin, vmax=vmax, extent=extent, cmap=cmap, aspect='auto', norm=norm,
                    interpolation=interpolation, **kwargs)

    # figure formatting

    # x axis
    ax.xaxis_date()
    ax.xaxis.set_major_locator(HourLocator(interval=2))
    ax.xaxis.set_major_formatter(DateFormatter('%H:%M'))
    ax.xaxis.grid(True)
    plt.xlabel('UTC Time')

    # y axis
    ax.yaxis_date()
    dmin, dmax = ax.yaxis.get_data_interval()
    number_of_days = (num2date(dmax) - num2date(dmin)).days
    # AutoDateLocator is not suited in case few data is available
    if abs(number_of_days) <= 35:
        ax.yaxis.set_major_locator(DayLocator())
    else:
        ax.yaxis.set_major_locator(AutoDateLocator())
    ax.yaxis.set_major_formatter(DateFormatter("%a, %d %b %Y"))

    # plot colorbar
    cbticks = np.logspace(np.log10(vmin), np.log10(vmax), 11, endpoint=True)
    cb = plt.colorbar(format='%.0f', ticks=cbticks)
    cb.set_label(cblabel)

    # plot title
    plt.title(title)

    return im


def fanchart(timeseries, **kwargs):
    """
    Draw a fan chart of the daily consumption profile.

    The fan chart shows the different quantiles of the daily consumption, with
    the blue line representing the median, and the black line the average.
    By default, the consumption of the whole day is taken, but one can select
    the hours of interest, e.g. night time standby consumption.

    Parameters
    ----------
    timeseries : pandas.Series
    start_hour, end_hour : int or float, optional
        Start and end hours of period of interest, default values are 0, 24
        As of now, ensure that start_hour < end_hour
    ylabel, title : str
        If not None, these determine the labels of y axis and/or title.
        If None, the name of the timeseries is used if defined.
    """

    start_hour = 2. + kwargs.pop('start_hour', 0.) / 24.
    end_hour = 2. + kwargs.pop('end_hour', 24.) / 24.
    ylabel = kwargs.pop('ylabel', timeseries.name if timeseries.name else '')
    title = kwargs.pop('title', 'carpet plot: ' + timeseries.name if timeseries.name else '')
    # data preparation
    if timeseries.dropna().empty:
        print('skipped {} - no data'.format(title))
        return
    ts = timeseries.resample('min', how='mean', label='left', closed='left')

    # convert to dataframe with date as index and time as columns by
    # first replacing the index by a MultiIndex
    # tz_convert('UTC'): workaround for https://github.com/matplotlib/matplotlib/issues/3896
    mpldatetimes = date2num(ts.index.tz_convert('UTC').astype(dt.datetime))
    ts.index = pd.MultiIndex.from_arrays(
        [np.floor(mpldatetimes), 2 + mpldatetimes % 1])  # '2 +': matplotlib bug workaround.
    # and then unstacking the second index level to columns
    df = ts.unstack()
    df = df.T.truncate(start_hour, end_hour)

    num = 20
    num_max = 4
    df_quant = df.quantile(np.linspace(0., 1., 2 * num + 1))

    # data plotting

    fig, ax = plt.subplots()
    im = plt.plot(df.columns, df_quant.iloc[num], 'b', label='median')
    for i in range(1, num):
        plt.fill_between(df.columns, df_quant.iloc[num - i], df_quant.iloc[min(num + i, 2 * num - num_max)], color='b',
                         alpha=0.05)
    plt.plot(df.columns, df.mean(), 'k--', label='mean')
    plt.legend()

    # x axis
    ax.xaxis_date()
    plt.xlim(df.columns[0], df.columns[-1])
    plt.ylabel(ylabel)

    # plot title
    plt.title(title)
    plt.grid(True)

    return im



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 21 15:31:36 2013 by Carlos Dierckxsens

"""
import sys
import datetime as dt
import pandas as pd
import requests
import os
import pytz
import re
import zipfile
import glob
import time

def pull_api(sensor, token, unit, interval='day', resolution='minute'):
   
    """     
   
    Function for downloading data from fluksometers
    
    Parameters
    ----------
    - interval : string specifying the interval (day, month, ...)
    - sensor :  sensor name (from the flukso.net sensor tab)
    - token :  sensor token (from the flukso.net sensor tab)
    - resolution : time resolution (e.g. minute, 15min, hour, day, week, month, 
      year, decade, night)
    - unit : unit of measurements (e.g. watt, kwhperyear, lperday)

    Note
    ----
    The Flukso Server will automatically restrict the data to what's available
    
    
    Returns
    -------
    Result of the http request with the raw data.  
    Use the save2csv function to parse and save.
    """
    
    payload = {'interval'   :   interval,
               'resolution' :   resolution,
               'unit'       :   unit}
               
    headers = {'Accept'     :   'application/json', 
               'X-Version'  :   '1.0', 
               'X-Token'    :   token}  
               
    url = 'https://api.flukso.net' + '/sensor/' + sensor 
    
    # Send Request
    try:    
        s = requests.Session()
        r = s.get(url, params=payload, headers=headers, verify=False)
    except:
        print("-------> Problem with HTTP request to Flukso <-------")
    
    # check variable
    if not r.ok:
        print("The flukso api GET request did not succeed.")
        print("Some details:")
        print("Request headers:")
        print(r.request.headers)
        print("Request url:")
        print(r.request.url)
    
    return r


def parse(r):
    """
    Parse and return a pandas TimeSeries object
    """
    
    
    # Create TimeSeries   
    try:
        d = {}
        for tup in r.json():
            d[dt.datetime.fromtimestamp(tup[0])] = tup[1]
        
    except:
        print("-------> Problem with Flukso data parsing <-------")
        raise
    
    #pdb.set_trace()
    ts = pd.TimeSeries(data=d)
    # Convert the index to a pandas DateTimeIndex 
    ts.index = pd.to_datetime(ts.index)
    
    return ts


def save_file(df, folder=None, file_type='csv', prefix=''):
    """
    Save the TimeSeries or DataFrame to csv or hdf with specified name
    
    Parameters
    ----------
    df : pandas Timeseries or Dataframe
    folder : folder, default=None
        Folder where the file will be saved. Defaults to os.getcwd()
    file_type : {'csv', 'hdf'}, default='csv'
        File type of the saved file.
    prefix = string, default=''
        Name prefix for the file name, usually FLxxxxxxx_sensorid
        
    Returns
    -------
    path : abspath
        Absolute path to the saved file
    """
    
   
    # save to file
    folder = folder or os.getcwd()
    s = df.index[0].strftime(format="%Y-%m-%d_%H-%M-%S")
    e = df.index[-1].strftime(format="%Y-%m-%d_%H-%M-%S")
    prefix = prefix + '_FROM_' + s + '_TO_' + e
    if file_type == 'csv':
        path = os.path.join(folder, prefix + '.csv')
        df.to_csv(path, header=False)
    elif file_type == 'hdf':
        path = os.path.join(folder, prefix + '.hdf')
        df.to_hdf(path, 'df', mode='w')
    else:
        raise Exception('fluksoapi.load: file_type should be either csv or hdf')
    return path

   
def load_file(path):
    """
    Load a previously saved csv or hdf file into a dataframe and return it.
    
    Parameters
    ----------
    path : path
        Path to a csv or hdf file.  Filename should be something like fluksoID_sensor_FROM_x_to_y.csv

    Returns
    -------
    df : pandas.DataFrame
        The dataframe will have a DatetimeIndex with UTC timezone.  The 
        column will be the sensor-ID, extracted from the csv filename. If invalid filename is given, an empty dataframe will be returned.
    
    """
    if len(path) == 0 or 'FROM' not in path:
        return pd.DataFrame()
    
    if '.csv' in path:
        df = pd.read_csv(path, index_col = 0, header=None, parse_dates=True)
        # Convert the index to a pandas DateTimeIndex 
        df.index = pd.to_datetime(df.index)
        df.index = df.index.tz_localize('UTC')
        df.columns = [path.split('_')[-7]]
    elif '.hdf' in path:
        df = pd.read_hdf(path, 'df')
    return df


def load_sensor(folder, sensor, dt_start=None, dt_end=None, files=None, error_no_files=True):
    """
    Load sensor
    
    Parameters
    ----------
    folder : path
        Folder containing the hdf and/or csv files
    sensor : hex
        Sensor for which files are to be consolidated
    dt_start, dt_end: datetime or equivalent, optional
    files : optional
        if not provided, sensor files are searched in the folder
    error_no_files : boolean, default True
        If True a ValueError is raised, if False an empty dataframe is returned
    
    Returns
    -------
    combination: dataframe
    """
    files = files or glob.glob(os.path.join(folder, '*' + sensor + '*'))
    if len(files) == 0:
        if error_no_files:
            # if no valid (unhidden) files are found, raise a ValueError.
            raise ValueError('No files found for sensor {} in {}'.format(sensor, folder))
        else:
            print('No files found for sensor {} in {}'.format(sensor, folder))
            return pd.DataFrame()
    print("About to combine {} files for sensor {}".format(len(files), sensor))
    dfs = [load_file(f) for f in files]
    combination = dfs[0]
    for df in dfs[1:]:
        combination = combination.combine_first(df)
    combination = combination.ix[dt_start:dt_end]
    return combination


def consolidate_sensor(folder, sensor, file_type='csv', dt_day=None, remove_temp=False):
    """
    Merge all csv and/or hdf files for     
    - the given sensor
    - and the given day
    into a single csv or hdf file
    
    Parameters
    ----------
    folder : path
        Folder containing the csv and/or hdf files
    sensor : hex
        Sensor for which files are to be consolidated
    file_type : {'csv', 'hdf'}, default='csv'
        File type of the saved file.
    dt_day : (optional) datetime
        If a valid datetime is passed, only files containing data from this day
        will be considered
    remove_temp : (optional) Boolean, default=False
        If True, only the resulting consolidated file is kept, the files that
        have been consolidated are deleted.
        
    Returns
    -------
    path : path of the consolidated csv or hdf file
    
    """
    
    # List all files (ABSPATH) for the given sensor in the given path, without hidden files			
    # glob.glob() is equivalent to os.listdir(folder) without the hidden files (start with '.') 
    # and returned as absolute paths
    files = glob.glob(os.path.join(folder, '*' + sensor + '*'))
    if dt_day is not None:
        #only retain data from this day
        dt_day_string = dt_day.strftime(format="%Y-%m-%d")
        files = [f for f in files if dt_day_string in f]
        dt_start = dt.datetime.strptime(dt_day_string, "%Y-%m-%d")
        dt_end = dt_start + dt.timedelta(days=1)
    else:
        dt_start = dt_end = None
    if len(files) == 0:
        # if no valid (unhidden) files are found, raise a ValueError.
        raise ValueError('No files found for sensor {} in {}'.format(sensor, folder))
    elif len(files) == 1:
        print("One file found and retained for sensor {}".format(sensor))
        return files[0]
    else:
        combination = load_sensor(folder, sensor, dt_start, dt_end, files)
        if remove_temp:
            for f in files:
                os.remove(os.path.join(folder, f))
            print("Removed the {} temporary files".format(len(files)))
        
        # Obtain the new filename prefix, something like FX12345678_sensorid
        # the _FROM....hdf will be added by the save_hdf method
        prefix = files[-1].split('_FROM')[0]
        path = save_file(combination, folder, file_type=file_type, prefix=prefix)
        print('Saved ', path)
        return path


def consolidate_folder(folder, file_type='csv'):
    
    sensor_set = {x.split('_')[-7] for x in glob.glob(os.path.join(folder, '*'))}
    print('About to consolidate {} sensors'.format(len(sensor_set)))
    for sensor in sensor_set:
        consolidate_sensor(folder, sensor, file_type=file_type, remove_temp=True)
    

def synchronize(folder, unzip=True, consolidate=True, file_type='hdf'):
    """Download the latest zip-files from the opengrid droplet, unzip and consolidate.
    
    The files will be stored in folder/zip and unzipped and 
    consolidated into folder/csv
    
    Parameters
    ----------
    
    folder : path
        The *data* folder, containing subfolders *zip* and *csv*
    unzip : [True]/False
        If True, unzip the downloaded files to folder/csv
    consolidate : [True]/False
        If True, all csv files in folder/csv will be consolidated to a 
        single file per sensor
    
    Notes
    -----
    
    This will only unzip the downloaded files and then consolidate all
    csv files in the csv folder.  If you want to rebuild the consolidated
    csv from all available data you can either delete all zip files and 
    run this function or run _unzip(folder, consolidate=True) on the 
    data folder.
        
    """
    t0 = time.time()
    if not os.path.exists(folder):
        raise IOError("Provide your path to the data folder where a zip and csv subfolder will be created.")
    from opengrid import config
    # Get the pwd; start from the path of this current file 
    c = config.Config()
    pwd = c.get('opengrid_server', 'password')
    host = c.get('opengrid_server','host')
    port = c.get('opengrid_server','port')
    user = c.get('opengrid_server','user')
    URL = "".join(['http://',host,':',port,'/'])
    
    # create a session to the private opengrid webserver
    session = requests.Session()
    session.auth = (user, pwd)
    resp = session.get(URL)
    
    # make a list of all zipfiles
    pattern = '("[0-9]{8}.zip")' 
    zipfiles = re.findall(pattern, resp.content)
    zipfiles = [x.strip('"') for x in zipfiles]
    zipfiles.append('all_data_till_20140711.zip')
    
    zipfolder = os.path.join(folder, 'zip')    
    csvfolder = os.path.join(folder, 'csv')

    # create the folders if they don't exist
    for fldr in [zipfolder, csvfolder]:
        if not os.path.exists(fldr):
            os.mkdir(fldr)
    
    downloadfiles = [] # these are the successfully downloaded files       
    for f in zipfiles:
        # download the file to zipfolder if it does not yet exist
        if not os.path.exists(os.path.join(zipfolder, f)):
            print("Downloading {}".format(f))       
            with open(os.path.join(zipfolder, f), 'wb') as handle:
                response = session.get('http://95.85.34.168:8080/' + f, stream=True)
        
                if not response.ok:
                    raise IOError('Something went wrong in downloading of {}'.format(f))
        
                for block in response.iter_content(1024):
                    if not block:
                        break
                    handle.write(block)
            downloadfiles.append(f)
            
    t1 = time.time()
    # Now unzip and/or consolidate
    if unzip:
        _unzip(folder, downloadfiles)
    t2 = time.time()
    if consolidate:
        consolidate_folder(csvfolder, file_type=file_type)
    t3 = time.time()
    print('Download time: {} s'.format(t1-t0))
    print('Unzip time: {} s'.format(t2-t1))
    print('Consolidate time: {} s'.format(t3-t2))
    print('Total time: {} s'.format(t3-t0))
        

def _unzip(folder, files='all'):
    """
    Unzip zip files from folder/zip to folder/csv
        
    Parameters
    ----------
    
    folder : path
        The *data* folder, containing subfolders *zip* and *csv*
    files = 'all' (default) or list of files
        Unzip only these files
    
    """

    zipfolder = os.path.join(folder, 'zip')    
    csvfolder = os.path.join(folder, 'csv')

    # create the folders if they don't exist
    for fldr in [zipfolder, csvfolder]:
        if not os.path.exists(fldr):
            os.mkdir(fldr)

    if files == 'all':
        files = os.listdir(zipfolder)
    print('About to unzip {} files'.format(len(files)))
    badfiles = []
    
    for f in files:
        # now unzip to zipfolder
        try:       
            z = zipfile.ZipFile(os.path.join(zipfolder, f), 'r')
            z.extractall(path=csvfolder)
        except:
            badfiles.append(f)
            pass
    
    if badfiles:
        print("Could not unzip these files:")
        for f in badfiles:
            print(f)
  
    
def update_tmpo(tmposession, hp):
    """
    Update the tmpo database with all sensors from a houseprint object.
    This does NOT synchronize the data, only loads the sensors.  
    
    Parameters
    ----------
    tmposession : tmpo.Session object
    hp : a Houseprint object
    
    Returns
    -------
    tmposession : return the tmpo.Session
    """
    
    # get a list of all sensors in the hp.  The list contains tuples (sensor,token)
    sensors_tokens = hp.get_all_sensors(tokens=True)
    for s,t in sensors_tokens:
        tmposession.add(s,t)

    print("This tmpo session was updated with in total {} sensors".format(len(sensors_tokens)))
    return tmposession
    
    
def load_tmpo(tmposession, sensors, start=None, end=None):
    """
    Load data from one or more sensors into a pandas DataFrame
    
    Parameters
    ----------
    tmposession : tmpo.Session object
        tmpo session
    sensors : Str or List
        String: single sensor to be loaded
        List: list of sensors to be loaded
    start, end : Datetime, float, int, string or pandas Timestamp (default=None)
        Anything that can be parsed into a pandas.Timestamp
        If start is None, load all available data
        If end is None, end is the current time
    
    Returns
    -------
    df : pandas DataFrame
        DataFrame with DatetimeIndex and sensor-ids as columname.  If only a 
        single sensor is given, return a DataFrame instead of a Timeseries.

    Raises
    ------
    If no data is found, do not return an empty dataframe but raise
    a ValueError.
    
    """

    if isinstance(sensors, str):
        sensors = [sensors]
    
    # convert start and end to epoch
    if start is None:
        startepoch = 0
    else:
        # use parse_date to convert to pd.Timestamp and from there to POSIX
        startepoch = _parse_date(start).value/1e9
        
        # convert start and end to epoch
    if end is None:
        endepoch = sys.maxint
    else:
        # use parse_date to convert to pd.Timestamp and from there to POSIX
        endepoch = _parse_date(end).value/1e9

    # get list of timeseries    
    dfs = []    
    for s in sensors:
        try:        
            ts = tmposession.series(sid=s, head=startepoch, tail=endepoch)
        except Exception as e:
            print("No tmpo data for sensor {}".format(s))
        else:
            if len(ts) > 0:
                dfs.append(ts)
    
    df = pd.concat(dfs, axis=1)
    # convert POSIX timestamp (seconds since epoch) to DatetimeIndex    
    df.index = pd.to_datetime((df.index.values*1e9).astype(int))    
    return df
    
    
def load(path_csv, sensors, start=None, end=None):
    """
    Load data from one or more sensors into a pandas DataFrame.  
    
    Parameters
    ----------
    path_csv : str
        Folder containing the csv files with data to be loaded
    sensors : str or list of str
        Sensors to be loaded
    start, end : Datetime, float, int, string or pandas Timestamp, optional
        Anything that can be parsed into a pandas.Timestamp
        If start and end are not provided, all available data is loaded.
    
    Returns
    -------
    df : pandas DataFrame
        DataFrame with DatetimeIndex and sensor ids as column names.

    Raises
    ------
    If no single sensor is found, do not return an empty dataframe but raise
    a ValueError. Not implemented.
    
    Notes
    -----
    Currently, this function only calls ``load_sensor`` (csv and/or hdf files).
    Ultimately, it will first call ``load_tmpo`` and if the tmpo database does
    not contain all historic data (depending on start), it will also call 
    ``load_sensor``. Not implemented.
    
    """
    if isinstance(sensors, str):
        sensors = [sensors]
    
    dataframes = [load_sensor(path_csv, sensor, start, end, error_no_files=False) for sensor in sensors]
    df = pd.concat(dataframes, axis=1)
    df.index = df.index.tz_convert(pytz.timezone('Europe/Brussels'))
    
    print('{} sensors loaded'.format(len(df.columns)))
    df = df.dropna(axis=1, how='all')
    print('{} sensors retained'.format(len(df.columns)))
    print("Size of dataframe: {}".format(df.shape))

    return df


def _parse_date(d):
    """
    Return a pandas.Timestamp if possible.  
    
    Parameters
    ----------
    d : Datetime, float, int, string or pandas Timestamp
        Anything that can be parsed into a pandas.Timestamp
        
    Returns
    -------
    pts : pandas.Timestamp
    
    Raises
    ------
    ValueError if it was not possible to create a pandas.Timestamp
    """
    
    if isinstance(d, float) or isinstance(d, int):
        # we have a POSIX timestamp IN SECONDS.
        pts = pd.Timestamp(d, unit='s')
        return pts
        
    try:
        pts = pd.Timestamp(d)
    except:
        raise ValueError("{} cannot be parsed into a pandas.Timestamp".format(d))
    else:
        return pts
